import functools
import os
import pandas as pd
import numpy as np
from pathlib import Path
from pyadlml.constants import AREA, START_TIME, END_TIME, DEVICE, VALUE, TIME, ACTIVITY, STRFTIME_PRECISE
from pyadlml.dataset._core.activities import ActivityDict
from pyadlml.dataset.util import extract_kwargs

DATA_NAME = 'devices.csv'
DEV_MAP_NAME = 'device_map.csv'
ACT_MAP_NAME = 'activity_map.csv'
DEV2AREA_NAME = 'device_area_map.csv'
ACT2AREA_NAME = 'activity_area_map.csv'
ACT_NAME = 'activities_subject_%s.csv'
MAP_ID = 'id'



def _folder2filename(filename):
    def deco(func):
        @extract_kwargs
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            path = kwargs.pop('path')
            path = Path(path) if isinstance(path, str) else path
            if not path.is_file() and path.is_dir():
                # If there is no file and the path is maps to a directory
                # create the filename after convention
                kwargs['path'] = path.joinpath(filename)
            else:
                kwargs['path'] = path       

            return func(*args, **kwargs)
        return wrapper
    return deco


@_folder2filename(ACT_NAME)
def read_activities(path: Path, act_map: dict = None):
    """
    """
    df_acts = pd.read_csv(path)

    if act_map is not None:
        df_acts[ACTIVITY] = df_acts[ACTIVITY].map(act_map)

    df_acts[START_TIME] = pd.to_datetime(df_acts[START_TIME], dayfirst=True)
    df_acts[END_TIME] = pd.to_datetime(df_acts[END_TIME], dayfirst=True)
    return df_acts

@_folder2filename(ACT_NAME)
def write_activities(df: pd.DataFrame, path) -> None:
    df.to_csv(path, index=False, date_format=STRFTIME_PRECISE)


@_folder2filename(DATA_NAME)
def write_devices(df, path):
    df.to_csv(path, index=False, date_format=STRFTIME_PRECISE)

@_folder2filename(DATA_NAME)
def read_devices(path: Path, dev_map: dict = None):
    """
    Parameters
    ----------
    path: Path
        either path to file direct, or path to folder conatining
        'devices.csv' file
    """
    df_devs = pd.read_csv(path)

    if dev_map is not None:
        df_devs[DEVICE] = df_devs[DEVICE].map(dev_map)

    # TODO refactor, ensure that True, False mapping is only applied
    #      to devices with two values 
    df_devs[VALUE] = df_devs[VALUE].astype(str).replace({
        "0": False, "0.0": False, 
        "1": True, "1.0": True,
        "True":True, "False": False,
        "true":True, "false": False,
    })
    df_devs[TIME] = pd.to_datetime(df_devs[TIME], dayfirst=True)
    df_devs = df_devs.reset_index(drop=True)
    return df_devs

@_folder2filename(DEV2AREA_NAME)
def read_device2area(path: Path, dev_map: dict) -> pd.DataFrame:
    """"""
    dev2area = pd.read_csv(path)
    dev2area[DEVICE] = dev2area[DEVICE].map(dev_map)
    assert (dev2area.columns == np.array([DEVICE, AREA])).all()
    return dev2area

@_folder2filename(DEV2AREA_NAME)
def write_device2area(path: Path, df: pd.DataFrame):
    assert (df.columns == np.array([DEVICE, AREA])).all()
    df.to_csv(path, sep=',', index=False)


@_folder2filename(ACT2AREA_NAME)
def read_activity2area(path: Path, act_map: dict) -> pd.DataFrame:
    """"""
    act2area = pd.read_csv(path)
    act2area[ACTIVITY] = act2area[ACTIVITY].map(act_map)
    assert (act2area.columns == np.array([ACTIVITY, AREA])).all()
    return act2area

@_folder2filename(ACT2AREA_NAME)
def write_activity2area(path: Path, df: pd.DataFrame):
    assert (df.columns == np.array([ACTIVITY, AREA])).all()
    df.to_csv(path, sep=',', index=False)


@_folder2filename(DEV_MAP_NAME)
def read_device_map(path: Path) -> dict: 
    return pd.read_csv(path, index_col=MAP_ID)\
                .to_dict()[DEVICE]

@_folder2filename(DEV_MAP_NAME)
def write_device_map(df: pd.DataFrame, path: Path):
    assert (df.columns == np.array(['id', DEVICE])).all()
    df.to_csv(path, sep=',', index=False)

@_folder2filename(ACT_MAP_NAME)
def read_activity_map(path: Path) -> dict:
    return pd.read_csv(path, index_col=MAP_ID)\
                .to_dict()[ACTIVITY]

@_folder2filename(ACT_MAP_NAME)
def write_activity_map(df: pd.DataFrame, path: Path):
    assert (df.columns == np.array(['id', ACTIVITY])).all()
    df.to_csv(path, sep=',', index=False)    



def load(folder_path: str, subjects:list=[], retain_corrections=False) -> dict:
    """
    Loads a dataset generated by activity-assistant from a specified folder.

    Parameters
    ----------
    folder_path : str
        The path to the folder where the dataset is located.

    subjects : list of str
        The subjecst names that are to be included. If the activity file
        is named *activities_subject_foo.csv*, then *foo* is the corresponding name.

    Examples
    --------
    The dataset ``my_dataset.zip`` is extracted in the folder ``/path/to/``.

    >>> from pyadlml.dataset import load_act_assist
    >>> data = load_act_assist('/path/to/my_dataset/', subjects=['chris'])

    Returns
    -------
    data : object
    """
    assert isinstance(folder_path, str) or isinstance(folder_path, Path)
    assert isinstance(subjects, list)
    fp = Path(folder_path) 


    # Get mappings
    act_map = read_activity_map(fp.joinpath(ACT_MAP_NAME))
    dev_map = read_device_map(fp.joinpath(DEV_MAP_NAME))

    # Get dev and act list
    lst_dev = dev_map.values()
    lst_act = act_map.values()

    # Get area mappings
    dev2area = read_device2area(fp.joinpath(DEV2AREA_NAME), dev_map)
    act2area = read_activity2area(fp.joinpath(ACT2AREA_NAME), act_map)


    df_dev = read_devices(fp.joinpath(DATA_NAME), dev_map)

    # If no subjects where specified read the files from the folder following the naming schemata
    if not subjects:
        for file in Path(fp).iterdir():
            if ACT_NAME[:18] in file.name:
                subjects.append(file.name[19:-4])

    act_dct = ActivityDict()
    for subject in subjects:
        df_act = read_activities(Path(fp).joinpath(ACT_NAME%(subject)),
                                  act_map)
        act_dct[subject] = df_act

    return dict(
        devices=df_dev,
        device_list=lst_dev,
        activities=act_dct,
        activity_list=lst_act,
        dev2area=dev2area,
        act2area=act2area
    )



from pyadlml.dataset.plot.matplotlib.act_and_devs import (
    plot_contingency_states as plot_contingency_states,
    contingency_events as plot_contingency_events,
    activities_and_device_states as plot_activities_and_states,
    activities_and_device_events as plot_activities_and_events
)

from pyadlml.dataset.plot.plotly.acts_and_devs import (
    contingency_states as plotly_contingency_states,
    contingency_events as plotly_contingency_events,
    activities_and_devices as plotly_activities_and_devices,
)

from pyadlml.dataset.plot.matplotlib.activities import (
    count as plot_activity_count,
    boxplot as plot_activity_boxplot,
    duration as plot_activity_duration,
    transitions as plot_activity_transitions,
    density as plot_activity_density,
    correction as plot_activity_correction
)

from pyadlml.dataset.plot.plotly.activities import (
    density as plotly_activity_density,
    bar_count as plotly_activity_bar_count,
    heatmap_transitions as plotly_activity_transitions,
    boxplot_duration as plotly_activity_boxplot_duration,
    bar_cum as plotly_activity_duration,
)

from pyadlml.dataset.plot.matplotlib.devices import (
    state_fractions as plot_device_state_fractions,
    state_boxplot as plot_device_state_boxplot,
    state_similarity as plot_device_state_cross_correlation,
    states as plot_device_states,
    inter_event_intervals as plot_device_inter_event_intervals,
    event_density_one_day as plot_device_event_density,
    event_count as plot_device_event_count,
    event_raster as plot_device_event_raster,
    event_cross_correlogram as plot_device_event_correlogram,
)

from pyadlml.dataset.plot.plotly.devices import (
    bar_count as plotly_device_event_count,
)

from pyadlml.dataset.plot.matplotlib.util import (
    plot_cv_impact_parameter,
)


from pyadlml.constants import TIME, VALUE
import numpy as np
import pandas as pd

from pyadlml.preprocessing import EventWindows


def get_tds_per_batch(X, window_size, stride, seq_type):
    """ returns the time deltas for each batch

    Parameters
    ----------
    X : pd.DataFrame
        A dataframe where at least one column has the name TIME. Can be either a `StateVector`
        or a `data.df_devices`.
    window_size : int
        The number of datapoints that are grouped together for a sequence
    stride : int
        The amount of steps that sliding window takes

    Returns
    -------
    df : pd.DataFrame
        A dataframe of the format (N) where for each batch the time differences
        are given
    """
    # extract time to index mapping
    time = X[TIME]
    index = X.index.values[:, np.newaxis]

    # print(index.shape)
    # slice and dice data
    ss = EventWindows(rep=seq_type, window_size=window_size, stride=stride)
    tmp, _ = ss.fit_transform(index)

    diffs = []
    for i in range(tmp.shape[0]):
        batch = time[tmp[i].squeeze()]

        # compute td difference to successor
        batch_shifted = batch.shift(-1)
        batch_diff = (batch_shifted - batch)[:-1]

        diffs.append(batch_diff)
    return diffs


def comp_tds_sums(X, window_size, stride, seq_type):
    sums = []
    diffs = get_tds_per_batch(X, window_size, stride, seq_type)
    for d in diffs:
        sums.append(d.sum())
    sums = np.array(sums)
    return sums


def comp_tds_sums_mean(X, window_size, stride, seq_type):
    sums = comp_tds_sums(X, window_size, stride, seq_type)
    return sums.mean()


def comp_tds_sums_median(X, window_size, stride):
    sums = comp_tds_sums(X, window_size, stride)
    return np.median(sums)


def df_density_binned(df, column_str, dt):
    """ Compute the binned density of samples

    Parameters
    ----------
    df : pd.DataFrame
        df of with columns [time, column_str]
    """

    df[TIME] = df[TIME].dt.floor(freq=dt).dt.time

    df[VALUE] = 1
    df = df.groupby([TIME, column_str]).sum().unstack()
    df = df.fillna(0)
    df.columns = df.columns.droplevel(0)

    # Add times bins where no sample hits anything
    range = pd.date_range(start='1/1/2000', end='1/2/2000', freq=dt).time
    df2 = pd.DataFrame(data=0, dtype=float, columns=df.columns, index=[d for d in range[:-1]])
    df2.update(df)
    df2 = df2.reset_index(drop=False).rename(columns={'index': TIME})
    return df2


import torch
import numpy as np
import pandas as pd
from sklearn.metrics import confusion_matrix
import plotly.graph_objects as go





def plotly_confusion_matrix(y_pred, y_true, labels, per_step=False):

    if not per_step:
        if isinstance(y_pred, torch.Tensor):
            y_pred = y_pred.detach().cpu().numpy()

        if isinstance(y_true, torch.Tensor):
            y_true = y_true.detach().cpu().numpy()

        cm = confusion_matrix(y_true=y_true, y_pred=y_pred).astype(int)


        fig = go.Figure(
            data=go.Heatmap(
                z=cm,
                x=labels,
                y=labels,
                hoverongaps=False),
        )
        fig.update_layout(
            title="Confusion Matrix", 
            xaxis=dict(title="Predicted"),
            yaxis=dict(title="True"),
        )

        return fig
    else:


        nr_steps = len(y_pred)

        if isinstance(y_pred[0], torch.Tensor):
            y_pred = [y.detach().cpu().numpy() for y in y_pred]

        if isinstance(y_true[0], torch.Tensor):
            y_true = [y.detach().cpu().numpy() for y in y_true]

        cms = [confusion_matrix(y_true=yt, y_pred=yp).astype(int) for (yp, yt) in zip(y_pred, y_true)]


        fig = go.Figure()
        import plotly.express as px

        for step in range(nr_steps):
            fig.add_trace(
                go.Heatmap(
                    name =f'CM_{step}',
                    z=cms[step],
                    x=labels,
                    y=labels,
                    colorscale='Viridis',
                    hovertemplate='Pred: %{x}<br>Truth: %{y}<br>Times: %{z} <extra></extra>',
                    hoverongaps=False
                ),
            )

        # Make last confusion matrix visible
        fig.data[nr_steps-1].visible = True

        steps = []
        for i in range(len(fig.data)):
            step = dict(
                method="update",
                args=[{"visible": [False] * len(fig.data)},
                    {"title": "Confusion matrix: " + str(i)}],  # layout attribute
            )
            step["args"][0]["visible"][i] = True  # Toggle i'th trace to "visible"
            steps.append(step)

        sliders = [dict(
            active=10,
            currentvalue={"prefix": "Epoch: "},
            pad={"t": 50},
            steps=steps
        )]

        fig.update_layout(
            sliders=sliders
        )

        # Share colorscale
        for i in range(nr_steps):
            fig.data[i]['coloraxis'] = 'coloraxis'

    return fig


from pyadlml.dataset._datasets.activity_assistant import \
    read_activities, write_activities, \
    read_devices, write_devices, \
    read_activity2area, write_activity2area, \
    read_device2area, write_device2area, \
    read_activity_map, write_activity_map, \
    read_device_map, write_device_map, \
    load as load_act_assist




import numpy as np
import pandas as pd
from pyadlml.constants import BOOL, CAT, DEVICE, END_TIME, START_TIME, TIME, VALUE
from pyadlml.dataset.util import get_first_states, get_last_states, infer_dtypes, str_to_timestamp


def remove_days(df_devices, df_activities, days=[], offsets=[], retain_corrections=False):
    """ Removes the given days from activities and devices and shifts the succeeding days by that amount
        forward.

    Parameters
    ----------
    df_devices : pd.DataFrame
    df_activities : pd.DataFrame
    days : list
        List of strings
    offsets : list
        Offsets that are added to the corresponding days.

    Returns
        df_devices : pd.DataFrame
        df_activities : pd.DataFrame
    """
    df_devs = df_devices.copy()
    df_acts = df_activities.copy()

    # Add offsets to the days specified
    for i in range(len(days)):
        days[i] = str_to_timestamp(days[i])
        if i < len(offsets):
            days[i] = days[i] + pd.Timedelta(offsets[i])

    # sort days from last to first
    days = np.array(days)[np.flip(np.argsort(days))]
    dtypes = infer_dtypes(df_devs)


    # 1. remove iteratively the latest day and shift the succeeding part accordingly
    for day in days:
        # when day is e.g 2008-03.23 00:00:00 then day after will be 2008-03-24 00:00:00
        # these variables have to be used as only timepoints can be compared as seen below
        day_after = day + pd.Timedelta('1D')

        # Remove devices events within the selected day
        mask = (day < df_devs[TIME]) & (df_devs[TIME] < day_after)
        removed_devs = df_devs[mask].copy()
        df_devs = df_devs[~mask]

        # shift the succeeding days timeindex one day in the past
        preceeding_days = (df_devs[TIME] < day)
        succeeding_days = (day_after < df_devs[TIME])

        succeeding_devs = df_devs[succeeding_days].copy()
        df_devs.loc[succeeding_days, TIME] = df_devs[TIME] - pd.Timedelta('1D')

        # Binary devices that change an odd amount of states in that day will have
        # a wrong state for the succeeding days until the next event
        #nr_dev_events = removed_devs[removed_devs[DEVICE].isin(dtypes[BOOL])]\
        #                            .groupby(by=[DEVICE]).count()
        #for dev in nr_dev_events.index:
        #    nr = nr_dev_events.loc[dev, TIME]
        #    if nr % 2 != 0 and (dev in dtypes[BOOL] or dev in dtypes[CAT]):
        #        print(f'Warning: Removed odd #events: device {dev} => inconsistent device states. Correction advised!')

        # Get first device states from after and last device states from earlier
        dev2last_state = get_last_states(df_devs[preceeding_days])
        dev2first_state = get_first_states(succeeding_devs)
        dev2last_removed_state = get_last_states(removed_devs)
        events_to_add = []
        eps = pd.Timedelta('1ms')
        for dev, val in dev2last_state.items():
            #if dev in dtypes[BOOL] and val == dev2first_state[dev]:
            #    # The case when an odd number of events appeared and therefore the states are
            #    # would be wrong 
            #    events_to_add.append([day+eps, dev, not dev2first_state[dev]])
            #    print(f'Warning: Detected odd #events: device {dev} => inconsistent device states. ')
            #    eps+= pd.Timedelta('1ms')
            def last_states_are_not_equal(dev):
                try:
                    return dev2last_state[dev] != dev2last_removed_state[dev]
                except KeyError:
                    # The case when no events are removed for that device
                    # => no event appeared in the remove d time frame
                    # => the states are equal
                    return False
            if (dev in dtypes[CAT]+dtypes[BOOL]) and last_states_are_not_equal(dev):
                events_to_add.append([day+eps, dev, dev2last_removed_state[dev]])
                print(f'Warning: Removed category: device {dev} => inconsistent device states. Correction advised!')
                eps += pd.Timedelta('1ms')
        
        df_devs = pd.concat([df_devs, 
                  pd.DataFrame(columns=[TIME, DEVICE, VALUE], data=events_to_add)])



        # Remove activities in that day that do not extend from the previous day into the selected
        # day or extend from the selected day into the next day
        mask_act_within_day = (day < df_acts[START_TIME]) & (df_acts[END_TIME] < day_after)
        df_acts = df_acts[~mask_act_within_day]

        # Special case where one activity starts before the selected day and ends after the selected day
        mask_special = (df_acts[START_TIME] < day) & (day_after < df_acts[END_TIME])
        if mask_special.any():
            # Adjust the ending by removing one day
            df_acts.loc[mask_special, END_TIME] = df_acts[END_TIME] - pd.Timedelta('1D')

        # Shift Activities that start in or after the selected day by one day
        succeeding_days = (day <= df_acts[START_TIME]) & (day_after < df_acts[END_TIME])
        df_acts.loc[succeeding_days, START_TIME] = df_acts[START_TIME] - pd.Timedelta('1D')
        df_acts.loc[succeeding_days, END_TIME] = df_acts[END_TIME] - pd.Timedelta('1D')
        df_acts['shifted'] = False
        df_acts.loc[succeeding_days, 'shifted'] = True

        # Special case where one activity starts before the selected day and ends inside the selected day
        # and there is no activity after the selected day
        #  | day_before | Sel day      | day after
        #     |-------------|
        #    I can't just move the ending one day before as this would reverse START_TIME and END_TIME order
        # -> The last activity ending is clipped to the start of the selected day
        # TODO has to be done before any activity is shifted
        #mask_last_true = pd.Series(np.zeros(len(df_acts), dtype=np.bool_))
        #mask_last_true.iat[-1] = True
        #mask_special = (df_acts[START_TIME] < day) & (df_acts[END_TIME] <= day_after) & mask_last_true
        #if mask_special.any():
        #    assert mask_special.sum() == 1
        #    df_acts.loc[mask_special, END_TIME] = day

        df_acts = df_acts.sort_values(by=START_TIME).reset_index(drop=True)
        # Merge activities from the day_before that overlap with the shifted activities from day after
        # there are 4 cases where overlaps into the respective days have to be handled
        #                           |   db  |   sd  |   da  |=>|   db  |   sd  |   da  |=>|   db  |   sd  |   da  |
        # 1. db overlaps da            |--------|    |~|          |--------|                 |-----|~|
        #                                                              |~|
        # 2. db intersect into  da      |------|     |~~~~|         |------|                   |----|~~~~|
        #                                                               |~~~~|
        # 3. da overlaps db             |-|   |~~~~~~~|            |-|                       |-|~~~|
        #                                                       |~~~~~~~|
        # 4. da intersect into db       |---|  |~~~~~|          |---|                        |---|~~~|
        #                                                         |~~~~~|
        # 5. da intersect into db         |----||~~~~~|            |-----|                 |---|~~~|
        #                                                      |~~~~~~|
        # case 1:
        # select activities that cross the boundary between days
        mask_db_into_sd = (df_acts[START_TIME] < day) & (df_acts[END_TIME] > day) & (df_acts[END_TIME] < day_after)
        idxs = np.where(mask_db_into_sd)[0]
        assert len(idxs) <= 2
        if len(idxs) == 2:
            # case 5: case when both activities extend into each days
            # clip both to midnight
            df_acts.iat[idxs[0], 2] = day
            df_acts.iat[idxs[1], 1] = day + pd.Timedelta('1ms')
        if len(idxs) == 1:
            idx_overlapping = idxs[0]

            # Check if the overlapping activity is part of the shifted or not
            if df_acts.iat[idx_overlapping, 3]:
                # Case when the shifted activities extend into the day before the selected day
                last_unshifted_act = df_acts.loc[(df_acts[END_TIME] < day), :]\
                                .copy().sort_values(by=END_TIME, ascending=True)\
                                .iloc[-1, :]

                # clip extending activities start_time to the end_time of the first shifted activity
                df_acts.iat[idx_overlapping, 0] = last_unshifted_act[END_TIME] + pd.Timedelta('1ms')
            else:
                # Case when the previous activities extends into the selected day
                first_shifted_act = df_acts.loc[(df_acts[START_TIME] > day) & (df_acts[END_TIME] < day_after), :]\
                                            .copy().sort_values(by=START_TIME, ascending=True)\
                                            .iloc[0, :]

                # clip extending activities end_time to the start of the first shifted activity
                df_acts.iat[idx_overlapping, 1] = first_shifted_act[START_TIME] - pd.Timedelta('1ms')

        df_acts = df_acts.drop(columns='shifted')

        from pyadlml.dataset._core.devices import correct_devices
        df_devs, corrections_dev = correct_devices(df_devs, retain_corrections)

        #assert not _is_activity_overlapping(df_acts)

    if retain_corrections:
        try:
            corrections_act
        except:
            corrections_act = []
        try:
            corrections_dev
        except:
            corrections_dev = []
        return df_devs, df_acts, corrections_act, corrections_dev
    else:
        return df_devs, df_acts




import pandas as pd

# compute a chi-squared dependence table
def rep2onoff(rep):
    df = rep.copy().iloc[:,:-1] # get raw without activities
    df = df.reset_index(drop=True)
    df = df.astype(int)

    for device in df.columns:
        mask1 = df[[device]] == 1
        col_off = device + ' Off'
        df[[col_off]] = df[[device]].astype(bool)
        df[[col_off]] = ~df[[col_off]]
        df[[col_off]] = df[[col_off]].astype(int)
        df = df.rename(columns={device: device + ' On'})
    df = df.astype(bool)
    return df

def chi2(X, ):
    # sklearn feature selection chi2
    from sklearn.feature_selection import chi2
    rep = raw
    df = rep2onoff(rep)
    chi2stats = chi2(df.values, raw['activity'])
    index = [item + ' On' for item in rep.columns[:-1]]
    index += [item + ' Off' for item in rep.columns[:-1]]
    df = pd.DataFrame(data={'p-value':chi2stats[1], 'chi2 stat': chi2stats[0]}, index=index)
    df = df.sort_index()
    return df


def mut_inf_01(rep):
    # sklearn feature selection chi2
    from sklearn.feature_selection import mutual_info_classif as mic
    rep = raw
    df = rep2onoff(rep)
    X, y = df.values, raw['activity']
    
    mistats = mic(X, y, n_neighbors=4)
    index = [item + ' On' for item in rep.columns[:-1]]
    index += [item + ' Off' for item in rep.columns[:-1]]
    
    df = pd.DataFrame(data={'mutual information':mistats}, index=index)
    df = df.sort_index()
    return df

from sklearn.feature_selection import mutual_info_classif as mic

def mut_inf(rep):
    col_name = 'Mutual Information'
    X, y = rep.values[:,:-1], rep.values[:,-1:].squeeze()
    mistats = mic(X, y, n_neighbors=4)
    df = pd.DataFrame(data={col_name : mistats}, index=rep.columns[:-1])
    df = df.sort_values(by=col_name, ascending=False)
    return df


import sys
import pathlib

from testing.test_preprocessing import TestPreprocessingBase

working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))
import unittest
from test_dataset_base import TestDatasetBase


from pyadlml.dataset import set_data_home
from pyadlml.dataset import ACTIVITY, DEVICE, END_TIME, START_TIME, VAL, TIME,\
    set_data_home, fetch_casas_aruba, clear_data_home
TEST_DATA_HOME = '/tmp/pyadlml_testing'

class TestCasasArubaDataset(TestDatasetBase):
    __test__ = True
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _setUp(self):
        self.data = fetch_casas_aruba(keep_original=True, cache=False)
        self.fetch_method = fetch_casas_aruba
        self.data_home = TEST_DATA_HOME
        self.df_activity_attrs = ['df_activities']
        self.lst_activity_attrs = ['lst_activities']
        self.num_activities = [11]
        self.len_activities = [6475]
        self.num_rec_acts = [11]

        self.len_devices = 1589824
        self.num_devices = 31
        self.num_rec_devs = 31
        self.num_rec_binary_devs = 31


class TestCasasArubaPreprocessing(TestPreprocessingBase):
    __test__ = True
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fetch_method = fetch_casas_aruba
        self.data_home = TEST_DATA_HOME
        self.df_activity_attrs = ['df_activities']
        self.lst_activity_attrs = ['lst_activities']

if __name__ == '__main__':
    unittest.main()


import numpy as np
import pandas as pd
from pyadlml.pipeline import Pipeline
from torch.utils.data import Dataset

class TorchDataset(Dataset):

    def __init__(self, X, y, transforms: Pipeline, transform_params={}, class_encoding=True):

        import torch
        X_t, Y_t = transforms.fit_transform(X.copy(), y.copy(), **transform_params)

       # Cast X_t and Y_t to numpy
        if isinstance(X_t, pd.DataFrame) or isinstance(X_t, pd.Series):
            X_t = X_t.values
        if isinstance(Y_t, pd.DataFrame) or isinstance(Y_t, pd.Series):
            Y_t = Y_t.values
        assert isinstance(X_t, np.ndarray) and isinstance(Y_t, np.ndarray)

        # Create Y_t class encoding
        self.classes_ = np.unique(Y_t.flatten())
 
        if class_encoding:
            self.int2class_ = {k:v for k, v in enumerate(self.classes_)}
            self.int2class_func = np.vectorize(self.int2class_.get)
            self.class2int_ = {v:k for k, v in self.int2class_.items()}
            self.class2int_func = np.vectorize(self.class2int_.get)
            Y_t = self.class2int_func(Y_t)

        # Compute class frequencies
        # Position 0 equal freq of class with idx zero
        count_per_class = np.bincount(Y_t.flatten())
        assert len(self.classes_) == len(count_per_class)
        self.class_weights_ = torch.tensor(
            count_per_class.sum()/(len(count_per_class)*count_per_class
        ),dtype=torch.float32)

        # Create torch tensor dataset and let torch infer datatype
        # What could go wrong
        self.Xtr = torch.tensor(X_t).to(torch.float32)
        self.Ytr = torch.tensor(Y_t).long()

    def __len__(self):
        return self.Xtr.shape[0]

    @property
    def class_weights(self):
        return self.class_weights_

    @property 
    def dct_class_weights(self):
        return {k:v for k, v in zip(self.class2int_,self.classes_)}


    @property
    def n_classes(self):
        return len(self.classes_)

    def int2class(self, i: int):
        return self.int2class_[i]

    def class2int(self, cls: str):
        return self.class2int_[cls]

    def __getitem__(self, idx):
        return self.Xtr[idx], self.Ytr[idx]
 


import unittest
import numpy as np
import os
from matplotlib import pyplot as plt
import copy, os

import pyhsmm
from matplotlib.figure import Figure
from pyhsmm.util.text import progprint_xrange

class TestController(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass


    def test_examples(self):
        np.seterr(divide='ignore') # these warnings are usually harmless for this code

        #  load_basic data
        T = 1000
        data = np.loadtxt(os.path.join(os.path.dirname(__file__),'example-data.txt'))[:T]

        #  posterior inference

        Nmax = 25

        obs_dim = data.shape[1]
        obs_hypparams = {'mu_0':np.zeros(obs_dim),
                        'sigma_0':np.eye(obs_dim),
                        'kappa_0':0.25,
                        'nu_0':obs_dim+2}
        dur_hypparams = {'alpha_0':2*30,
                         'beta_0':2}

        obs_distns = [pyhsmm.distributions.Gaussian(**obs_hypparams) for state in range(Nmax)]
        dur_distns = [pyhsmm.distributions.PoissonDuration(**dur_hypparams) for state in range(Nmax)]

        hsmm = pyhsmm.models.WeakLimitHDPHSMM(
                alpha=6.,gamma=6., # these can matter; see concentration-resampling.py
                init_state_concentration=6., # pretty inconsequential
                obs_distns=obs_distns,
                dur_distns=dur_distns)
        hsmm.add_data(data,trunc=60) # duration truncation speeds things up when it's possible

        for idx in progprint_xrange(150):
           hsmm.resample_model()

        hsmm.plot()

        plt.show()


    def test_multinomials(self):
        T = 10
        data = np.loadtxt(os.path.join(os.path.dirname(__file__),'example-data-bin.txt'))[:T]
        print('data: ', data)
        mult_hyperparams = {
            'K': 3,
            'alpha_v0': ,
        }
        mult = pyhsmm.basic.distributions.Categorical(**mult_hyperparams)
        ll = mult.log_likelihood(data)
        print(ll)

    def test_examples_2(self):
        np.seterr(divide='ignore') # these warnings are usually harmless for this code

        SAVE_FIGURES = False

        #  load_basic data

        T = 1000
        data = np.loadtxt(os.path.join(os.path.dirname(__file__),'example-data.txt'))[:T]

        #  posterior inference

        # Set the weak limit truncation level
        Nmax = 25

        # and some hyperparameters
        obs_dim = data.shape[1]
        obs_hypparams = {'mu_0':np.zeros(obs_dim),
                        'sigma_0':np.eye(obs_dim),
                        'kappa_0':0.25,
                        'nu_0':obs_dim+2}
        dur_hypparams = {'alpha_0':2*30,
                         'beta_0':2}

        obs_distns = [pyhsmm.distributions.Gaussian(**obs_hypparams) for state in range(Nmax)]
        dur_distns = [pyhsmm.distributions.PoissonDuration(**dur_hypparams) for state in range(Nmax)]


        hsmm = pyhsmm.models.HSMM(
            alpha=6.,
            init_state_concentration=6.,
            obs_distns=obs_distns,
            dur_distns=dur_distns

        )
        #        alpha=6.,gamma=6., # these can matter; see concentration-resampling.py
        #        init_state_concentration=6., # pretty inconsequential
        #        obs_distns=obs_distns,
        #        dur_distns=dur_distns)
        hsmm.add_data(data, trunc=60) # duration truncation speeds things up when it's possible

        #for idx in progprint_xrange(150):
        #   posteriormodel.resample_model()
        #print(hsmm.predict())
        fig = hsmm.make_figure() # type: Figure
        print('figtype: ', type(fig))
        print('fig: ', fig)
        fig.savefig('test3.png')



        #Jposteriormodel.plot()
        print(hsmm)
        print(hsmm.log_likelihood(data))
        print(hsmm.states_list)
        print(hsmm.durations)

        #plt.show()



    def test_examples_binomial(self):
        np.seterr(divide='ignore') # these warnings are usually harmless for this code

        SAVE_FIGURES = False

        #  load_basic data

        T = 1000
        data = np.loadtxt(os.path.join(os.path.dirname(__file__),'example-data-bin.txt'))[:T]

        #  posterior inference

        # number of states
        N = 3
        # number of observations
        D = 3

        # and some hyperparameters
        weights = np.array([0.3,0.5,0.2])
        obs_hypparams = {'weights': weights, 'K' : N}#, 'K' : N}
                         #'alpha_0': 2*30,
                         #'alphav_0': 2*30}
        dur_hypparams = {'alpha_0':2*30,
                         'beta_0':2}
        trans_matrix = np.array([[0.3,0.5,0.2],
                                 [0.3,0.6,0.1],
                                 [0.1,0.4,0.5]])
        #trans_matrix = np.log(np.array([[0.3,0.5,0.2],
        #                 [0.3,0.6,0.1],
        #                 [0.1,0.4,0.5]]))

        obs_hypparams = {'mu_0':np.zeros(D),
                        'sigma_0':np.eye(D),
                        'kappa_0':0.25,
                        'nu_0':D+2}
        dur_hypparams = {'alpha_0':2*30,
                         'beta_0':2}

        obs_distns = [pyhsmm.distributions.Gaussian(**obs_hypparams) for state in range(N)]

        #print(trans_matrix)
        #print(isinstance(trans_matrix, np.ndarray))
        #print(trans_matrix.ndim)
        #print(trans_matrix.shape[1])
        #obs_distns = [pyhsmm.distributions.Multinomial(**obs_hypparams) for state in range(N)]
        dur_distns = [pyhsmm.distributions.PoissonDuration(**dur_hypparams) for state in range(N)]
        #dur_distns = [pyhsmm.distributions.B(**dur_hypparams) for state in range(N)]

        hsmm = pyhsmm.models.HSMM(
            alpha=6.,
            init_state_concentration=6.,
            obs_distns=obs_distns,
            dur_distns=dur_distns

        )

        #hsmm = pyhsmm.models.HSMM(
        #    #init_state_concentration=
        #    #alpha=
        #    #pi_0=
        #    #trans_distn=
        #    #alpha_b_0=
        #    #beta_b0=
        #    trans_matrix=trans_matrix,
        #    init_state_distn='uniform',
        #    obs_distns=obs_distns,
        #    dur_distns=dur_distns

        #)
        #        alpha=6.,gamma=6., # these can matter; see concentration-resampling.py
        #        init_state_concentration=6., # pretty inconsequential
        #        obs_distns=obs_distns,
        #        dur_distns=dur_distns)
        #hsmm.add_data(data, trunc=60) # duration truncation speeds things up when it's possible
        ll_prev = hsmm.log_likelihood(data)
        #print(ll_prev)
        #exit(-1)
        #hsmm.EM_fit()
        #for idx in progprint_xrange(150):
        #   hsmm.resample_model()
        #print(hsmm.predict())


        #Jposteriormodel.plot()
        #print(hsmm)
        #print(hsmm.log_likelihood(data))
        #print(hsmm.states_list)
        #print(hsmm.durations)

        #


    def test_hssmm(self):
        import numpy as np
        from matplotlib import pyplot as plt

        from pyhsmm.models import HSMMIntNegBinVariant
        from pyhsmm.basic.models import MixtureDistribution
        from pyhsmm.basic.distributions import Gaussian, NegativeBinomialIntegerRVariantDuration
        from pyhsmm.util.text import progprint_xrange

        #############################
        #  generate synthetic data  #
        #############################

        states_in_hsmm = 5
        components_per_GMM = 3
        component_hyperparameters = dict(mu_0=np.zeros(2), sigma_0=np.eye(2), kappa_0=0.01, nu_0=3)

        GMMs = [MixtureDistribution(
            alpha_0=4.,
            components=[Gaussian(**component_hyperparameters) for i in range(components_per_GMM)])
            for state in range(states_in_hsmm)]

        true_dur_distns = [
            NegativeBinomialIntegerRVariantDuration(np.r_[0., 0, 0, 0, 0, 1, 1, 1], alpha_0=5., beta_0=5.)
            for state in range(states_in_hsmm)]

        truemodel = HSMMIntNegBinVariant(
            init_state_concentration=10.,
            alpha=6., gamma=6.,
            obs_distns=GMMs,
            dur_distns=true_dur_distns)

        training_datas = [truemodel.generate(1000)[0] for i in range(5)]
        test_data = truemodel.generate(5000)[0]

        #####################################
        #  set up FrozenMixture components  #
        #####################################

        # list of all Gaussians
        component_library = [c for m in GMMs for c in m.components]
        library_size = len(component_library)

        # initialize weights to indicator on one component
        init_weights = np.eye(library_size)

        #obs_distns = [FrozenMixtureDistribution(
        #    components=component_library,
        #    alpha_0=4,
        #    weights=row)
        #    for row in init_weights]

        ################
        #  build HSMM  #
        ################

        dur_distns = [NegativeBinomialIntegerRVariantDuration(np.r_[0., 0, 0, 0, 0, 1, 1, 1], alpha_0=5., beta_0=5.)
                      for state in range(library_size)]

        #model = LibraryHSMMIntNegBinVariant(
        #    init_state_concentration=10.,
        #    alpha=6., gamma=6.,
        #    obs_distns=obs_distns,
        #    dur_distns=dur_distns)

        #for data in training_datas:
        #    model.add_data(data, left_censoring=True)
        #    # model.add_data_parallel(data,left_censoring=True)

        ###################
        ##  infer things  #
        ###################

        #train_likes = []
        #test_likes = []

        #for i in progprint_xrange(50):
        #    model.resample_model()
        #    # model.resample_model_parallel()
        #    train_likes.append(model.log_likelihood())
        #    # test_likes.append(model.log_likelihood(test_data,left_censoring=True))

        #print
        #'training data likelihood when in the model: %g' % model.log_likelihood()
        #print
        #'training data likelihood passed in externally: %g' % sum(
        #    model.log_likelihood(data, left_censoring=True) for data in training_datas)

        #plt.figure()
        #truemodel.plot()
        #plt.gcf().suptitle('truth')

        #plt.figure()
        #model.plot()
        #plt.gcf().suptitle('inferred')

        ## plt.figure()
        ## plt.plot(train_likes,label='training')
        ## plt.plot(test_likes,label='test')
        ## plt.legend()

        #plt.show()




from pyadlml.dataset.plot.matplotlib.util import func_formatter_seconds2time, func_formatter_seconds2time_log
from pyadlml.pipeline import XOrYTransformer, XAndYTransformer, YTransformer
from pyadlml.constants import END_TIME, OTHER, START_TIME, TIME
from sklearn.base import BaseEstimator, TransformerMixin
import pandas as pd
import numpy as np
from plotly.subplots import make_subplots

class Windows():
    REP_M2M = 'many-to-many'
    REP_M2O = 'many-to-one'

    def __init__(self, rep, window_size, stride):
        self.rep = rep
        self.window_size = window_size
        self.stride = stride



class ExplicitWindows():
    """
    https://www.mdpi.com/1424-8220/21/18/6037

    """
    pass


class TimeWindows(Windows):
    """ Divide data stream into time segments with a regular interval.

    Note
    ----
    Also works with regular sampled data (dt='xs').
    Van Kasteren et al. recommend 60s. 

    """
    def __init__(self, window_size : str, stride: str = None, rep: str ='many-to-many',  drop_empty_intervals=True):
        TransformerMixin.__init__(self)
        XOrYTransformer.__init__(self)
        Windows.__init__(self, rep, window_size, stride)

        self.drop_empty_intervals = drop_empty_intervals

    def fit_transform(self, X, y):
        return self.transform(X, y)

    @XOrYTransformer.x_or_y_transform
    def transform(self, X=None, y=None) -> np.ndarray:
        # Bring params into right format
        assert self.rep in [self.REP_M2M, self.REP_M2O]
        self.window_size = pd.Timedelta(self.window_size)
        self.stride = pd.Timedelta(self.stride) if self.stride is not None else self.window_size


        # TODO refactor add conversion for different input types
        assert isinstance(X, pd.DataFrame) or X is None
        assert isinstance(y, pd.DataFrame) or y is None

        # Solution without stride
        df = X.copy().sort_values(by=TIME)

        res_X = []

        res_y = [] if y is not None else None

        st = X[TIME].iloc[0]
        et = X[TIME].iloc[-1]
        win_st = st

        while win_st+self.window_size < et:
            win = (win_st, win_st + self.window_size)
            event_idxs = df[(win[0] < df[TIME]) & (df[TIME] < win[1])].index
            
            if not self.drop_empty_intervals or not event_idxs.empty:
                res_X.append(X.iloc[event_idxs].copy().values)
                if y is not None:
                    res_y.append(y.iloc[event_idxs].copy().values)

            win_st += self.stride

        return res_X, res_y 

class FuzzyTimeWindows():
    """
    """
    pass



class EventWindows(BaseEstimator, TransformerMixin, XOrYTransformer, Windows):
    """ Generate subsequences from a 

    .. image:: ../_static/images/many_to_many.svg
       :height: 200px
       :width: 500 px
       :scale: 90%
       :alt: alternate text
       :align: center


    .. image:: ../_static/images/reps/image.svg
       :height: 200px
       :width: 500 px
       :scale: 80%
       :alt: alternate text
       :align: center


    Many-To-Many
    ^^^^^^^^^^^^

    To get *many-to-many* batches use the window size :math:`w` to split the data

    .. math::
        f(X_{N,K},y_{N}) \rightarrow (X_{W, N,K}, y_{W, N})

    In addition you can specify a stride.


    .. code:: python

        from pyadlml.dataset import fetch_kasteren
        from pyadlml.preprocessing import EventWindows
        data = fetch_kasteren()

        raw = StateVectorEncoder(encode='raw', t_res='10s')\
              .fit_transform(data.df_devices)
        labels = LabelEncoder().fit_transform(raw, data.df_activities)

        X = raw.values
        y = labels.values

        X, y = EventWindows(rep='many-to-many', window_length=10, stride=2)\
               .fit_transform(X, y)

    Many-To-One
    ^^^^^^^^^^^

    .. math::
        f(X_{N,K},y_{N}) \rightarrow (X_{W, N, K}, y_{N})


    .. code:: python

        from pyadlml.preprocessing import SequenceDicer

        raw = StateVectorEncoder(encode='raw', t_res='10s')\
              .fit_transform(data.df_devices)
        labels = LabelEncoder().fit_transform(raw, data.df_activities)

        X = raw.values
        y = labels.values

        X, y = SequenceDicer(rep='many-to-one', window_length='20s')\
               .fit_transform(X, y)

    """
    def __init__(self, rep: str ='many-to-many', window_size: int =10, stride: int=1):
        """
        Parameters
        ----------
        rep: str 
        window_size: int
        stride: int

        """

        TransformerMixin.__init__(self)
        XOrYTransformer.__init__(self)
        Windows.__init__(self, rep, window_size, stride)


    def fit_transform(self, X, y=None):
        """
        Parameters
        ----------
        X : nd.array
            Some kind of numpy array
        y : nd.array

        Returns
        -------
        x : nd.array
            todo
        y : nd.array
            todo
        """
        return self.transform(X, y)

    @XOrYTransformer.x_or_y_transform
    def transform(self, X=None, y=None):
        """

        """
        assert self.rep in [self.REP_M2M, self.REP_M2O]

        if X is not None:
            assert self.window_size < len(X)
            assert self.stride < len(X)

            X = self._transform_X(X)

        if y is not None:
            assert self.window_size < len(y)
            assert self.stride < len(y)

            y = self._transform_Y(y)

        return X, y

    def _calc_new_N(self, n_old):
        return int(np.floor((n_old-self.window_size)/self.stride)+1)

    def _transform_X(self, X):
        """

        """
        if isinstance(X, pd.DataFrame) or isinstance(X, pd.Series):
            X = X.to_numpy()

        n_prime = self._calc_new_N(X.shape[0])
        new_shape =[n_prime, self.window_size, *X.shape[1:]]

        res = np.zeros(shape=new_shape, dtype=X.dtype)

        for r, i in enumerate(range(0, n_prime*self.stride, self.stride)):
            res[r] = X[i:i+self.window_size]
            #res[r, :, :] = X[i:i+self.window_size, :]

        return res.squeeze()

    def _transform_Y(self, y):
        """


        """

        if isinstance(y, pd.DataFrame):
            y = y.to_numpy().squeeze()
        elif isinstance(y, pd.Series):
            y = y.to_numpy()

        if self.rep == self.REP_M2M:
            n_prime = self._calc_new_N(y.shape[0])
            new_shape = [n_prime, self.window_size, *y.shape[1:]]

            res = np.zeros(shape=new_shape, dtype=y.dtype)

            for r, i in enumerate(range(0, n_prime*self.stride, self.stride)):
                res[r, :] = y[i: i+self.window_size]

            return res.squeeze()

        elif self.rep == self.REP_M2O:
            res = y[np.arange(self.window_size-1, y.shape[0], self.stride)]
            return res.squeeze()



    def plotly_windows_into_acts_and_devs(self, df_devs, df_acts):
        """

        """

        # Get times 
        Xt = self.fit_transform(df_devs) # (S, T, 3)


        from pyadlml.plot import plotly_activities_and_devices
        fig = plotly_activities_and_devices(df_devs, df_acts)

        # Plot times as vertical bars into acts_and_devs plot 
        for i in range(Xt.shape[0]):
            st = Xt[i][0, 0]
            et = Xt[i][-1, 0]
            fig.add_vline(x=st, line_width=2, line_color="Red", line_dash="dash")
            fig.add_vline(x=et, line_width=2, line_color="Red")

        return fig

    def plot_activities_per_window(self, X, y):
        """ For an already encoded stream count how many different
            activities are present in windows
        """

        raise NotImplementedError

    def plot_timelengths_per_window(self, X, y, times=None):
        """ Plot time length distribution over given encoded stream
        """
        Xt = self.transform(X)
        raise NotImplementedError
       

    @classmethod
    def plot_winsize_vs_activities(cls, df_devs, df_acts, window_sizes=None):
        """


        """
        window_sizes = np.arange(2, 10000, 1000) if window_sizes is None else window_sizes
        x_times = df_devs[TIME].values

        title = 'Window size vs. Activities'
        time_unit = 's'
        _ = None


        total_counts = [] 
        means = np.zeros(len(window_sizes))
        stds = np.zeros(len(window_sizes))
        med = np.zeros(len(window_sizes))
        for j, s in enumerate(window_sizes):
            Xt = cls(window_size=s, stride=s//2).fit_transform(x_times)

            # Get normalized time lengths
            counts = np.zeros(Xt.shape[0])
            for i in range(Xt.shape[0]):
                st, et = Xt[i, 0], Xt[i, 1]
                x_in_int = lambda x: (st < x) & (x < et)
                act_covers = lambda s, e: (df_acts[START_TIME] < s) & (e < df_acts[END_TIME])
                counts[i] = (x_in_int(df_acts[START_TIME]) | x_in_int(df_acts[END_TIME]) | act_covers(st, et)).sum()
            total_counts.append(counts)
            means[j] = counts.mean()
            stds[j] = counts.std()
            med[j] = np.median(counts)



        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        ax.plot(window_sizes, means, label='mean') 
        ax.plot(window_sizes, med, label='median')
        ax.plot(window_sizes, means + stds, linestyle='dashed', label='$\sigma$')
        ax.plot(window_sizes, means - stds, linestyle='dashed', label='-$\sigma$')
        ax.legend()
        return fig


    @classmethod
    def plotly_winsize_vs_length(cls, X, y, window_sizes=None, z_scale='linear', y_scale='linear'):
        """
        """
        import plotly.graph_objects as go
        window_sizes = np.arange(2, 1000, 10) if window_sizes is None else window_sizes

        assert window_sizes[0] > 1, 'Length is not defined for window containing one timestamp'

        # Reduce X to only time dimension
        # TODO, refactor 
        x_times = X[TIME].values

        title = 'Window size vs. window_length'
        cbarlabel = 'length' if z_scale == 'linear' else 'log length'
        n_bins = 100
        time_unit = 's'
        _ = None


        z = np.zeros((n_bins, len(window_sizes)), dtype=np.float32)
        dur_list = []
        dur_min = float('inf') 
        dur_max = float('-inf')
        for s in window_sizes:
            Xt = cls(window_size=s).fit_transform(x_times)
            # Get normalized time lengths
            Xt_dur = (Xt[:,-1] - Xt[:,0])/np.timedelta64(1, time_unit)
            
            dur_min = min(dur_min, Xt_dur.min())
            dur_max = max(dur_max, Xt_dur.max())
            dur_list.append(Xt_dur)

        # Make equal bin size from max to min
        # If log, Bins are equally sized in log space 
        #bin_gen = np.logspace if y_scale == 'log' else np.linspace
        #if y_scale == 'log':
        #    dur_min, dur_max = np.log10(dur_min), np.log10(dur_max)
        bin_gen = np.linspace
        bins = bin_gen(dur_min, dur_max, n_bins+1)


        for pos, d in enumerate(dur_list):
            hist, edges = np.histogram(d, bins=bins)
            z[:, pos] = hist

        z = np.log(z) if z_scale == 'log' else z

        fig = go.Figure(data=go.Heatmap(
                            x=window_sizes,
                            z=z,
                            colorscale='Viridis',
        ))
        if z_scale == 'log':
            tmpltstr = '%{x} event-length<br>' \
                     + ' -> %{customdata[1]} windows durating %{customdata[0]}.<extra></extra>'
            fig.data[0].hovertemplate = tmpltstr
            y_fmt = bins[:-1].copy().astype(object)

            for i in range(len(y_fmt)):
                if y_scale == 'log':
                    y_fmt[i] = func_formatter_seconds2time_log(y_fmt[i], _)
                else:
                    y_fmt[i] = func_formatter_seconds2time(y_fmt[i], _)

            y_fmt = np.repeat(y_fmt, len(window_sizes)).reshape(len(y_fmt), len(window_sizes))

            tmp = (np.exp(z.copy()) if z_scale == 'log' else z).astype(np.int32) 

            cd = np.array([y_fmt, tmp])
            fig['data'][0]['customdata'] = np.moveaxis(cd, 0, -1)

        #fig.update_yaxes(type='log')
        fig.update_layout(yaxis=dict(tickmode='array', tickvals=np.arange(len(bins[:-1])), ticktext=edges[:-1]))

        from pyadlml.dataset.plot.plotly.activities import _set_compact_title
        _set_compact_title(fig, title=title)
        return fig


    def get_y_times(self, times):
        times = self._transform_Y(times)

        if self.rep == self.REP_M2M:
            # TODO take only last y-obs per sequence??? 
            # Or average y for each position when plotting 
            raise NotImplementedError
        
        return times


import math

from hassbrain_algorithm.models._model import Model
import autograd.numpy as np
import autograd.numpy.random as npr
from scipy.stats import nbinom
import matplotlib.pyplot as plt
import ssm
from ssm.util import rle, find_permutation
from hassbrain_algorithm.models.util import create_xset_onebitflip

class BHSMM(Model):

    def __init__(self, controller):
        #self._cm = controller # type: Controller
        # training parameters
        self._training_steps = 50
        self._epsilon = None
        self._use_q_fct = False

        Model.__init__(self, "test", controller)

    def __str__(self):
        if self._hsmm is None:
            return "hmm has to be inits"
        else:
            s = "--"*10 + "\n"
            s += "States: " + str(self._hsmm.K) + "\n"
            s += "Obs:\t " + str(self._hsmm.D) + "\n"
            s += "pi:\t " + str(self._hsmm.init_state_distn) + "\n"
            s += "Trans:\t " + str(self._hsmm.transitions) + "\n"
            s += "M:\t " + str(self._hsmm.M) + "\n"
            return s

    def _model_init(self, dataset):
        K = len(self._state_list)       # number of discrete states
        D = len(self._observation_list) # dimension of the observation
        self._hsmm = ssm.HSMM(K, D, observations='bernoulli')

    def set_training_steps(self,  val):
        assert val > 1 and isinstance(val, int)
        self._training_steps = val


    def _sample(self, n, obs_seq=None):
        tmp1, tmp2 = self._hsmm.sample(T=n)
        return tmp1, tmp2

    def save_visualization(self, path_to_file):
        """
        save a visualization of the model to the given filepath
        :param path_to_file:
        :return:
        """
        pass

    def save_visualization_helper_decode_labels(self, label):
        """
        :return:
        """
        return self.decode_state_lbl(label)


    def predict_state_sequence(self, test_y):
        """

        :param test_y:
        :return:
        """
        pred_x = self._hsmm.most_likely_states(test_y)
        return pred_x

    def predict_obs_sequence(self, test_y):
        _, pred_y = self._hsmm.sample(test_y)
        return pred_y

    def draw(self, act_retrieval_meth):
        self._hmm.plot()
        return self._hsmm.generate_graphviz_dot_ext_lbl(act_retrieval_meth)
        #vg.render('test.gv', view=True)

    def _train_loss_callback(self, hmm, loss, *args):
        # todo in log models convert to normal Probability
        """
        :param hmm:
        :param loss:
        :param args:
        :return:
        """
        # this is due to the the loss param is actually the likelihood of the P(X|Model)
        # therefore the loss can be 1 - P(X|Model)
        loss = 1-loss
        self._bench.train_loss_callback(hmm, loss)

    def train(self, dataset):
        y_train = dataset.get_train_data()
        hsmm_em_lls = self._hsmm.fit(
            y_train,
            method="em",
            num_em_iters=self._training_steps)

        test_x, test_y = dataset.get_all_labeled_data()
        self.assign_states(test_x, test_y)
        return hsmm_em_lls

    def assign_states(self, true_z, true_y):
        """
        assigns the unordered hidden states of the trained model (on true_y)
        to the most probable state labels in alignment of true_z
        :param true_z
            the true state sequence of a labeled dataset
        :param true_y
            the true corresp. observation sequence of a labeled dataset

        assign
           z = true state seq [1,2,1,....,]
           tmp3 = pred. state seq [3,4,1,2,...,]
           match each row to different column in such a way that corresp
           sum is minimized
           select n el of C, so that there is exactly one el.  in each row
           and one in each col. with min corresp. costs


        match states [1,2,...,] of of the
        :return:
            None
        """
        # Plot the true and inferred states
        tmp1 = self._hsmm.most_likely_states(true_y)
        # todo temporary cast to int64 remove for space saving solution
        # todo as the amount of states only in range [0, 30
        true_z = true_z.astype(np.int64)
        tmp2 = find_permutation(true_z, tmp1)
        self._hsmm.permute(tmp2)

#-------------------------------------------------------------------
    # RT Node stuff

    def _classify(self, obs_seq):
        """
        get the most probable state/activity
        :param obs_seq:
        :return:
        """
        assert len(obs_seq[0]) == self._hsmm.D
        state_seq = self._hsmm.most_likely_states(obs_seq)
        pred_state = state_seq[-1:][0]
        return pred_state

    def _classify_multi(self, obs_seq):
        """
        calculates the probability of being in state z after observing obs_seq
        :param obs_seq:
            np array 2D
            e.g [ [0,1,0,0,1,...,0] , [0,0,1,0,...,1], ... ]
        :return:
            score_dict (dict)
                {0: 0.008754, 1: 0.1481, ... }
        """
        tmp = self._hsmm.filter(obs_seq)
        last_alpha = tmp[-1:][0]

        assert math.isclose(last_alpha.sum(), 1.)
        K = self._hsmm.K

        score_dict = {}
        for i in range(K):
            label = i
            score = last_alpha[i]
            score_dict[label] = score
        return score_dict

    def can_predict_next_obs(self):
        return True

    def can_predict_prob_devices(self):
        return True

    def _predict_next_obs(self, obs_seq):
        """
        returns index of most probable next observation
        """
        samples = create_xset_onebitflip(obs_seq[-1:][0])
        log_prob_states = self._hsmm.predict_xnp1(data=obs_seq,x=samples)
        # get max of pre_states
        max_idx = np.argmax(log_prob_states)
        if max_idx == len(log_prob_states) -1:
            return np.argmax(log_prob_states[:len(log_prob_states)-1])
        else:
            return max_idx

    def _predict_prob_xnp1(self, obs_seq):
        samples = create_xset_onebitflip(obs_seq[-1:][0])
        log_prob_states = self._hsmm.predict_xnp1(data=obs_seq,x=samples)
        normalizer = ssm.util.logsumexp(log_prob_states)
        norm_log_prob = log_prob_states - normalizer
        norm_log_prob = norm_log_prob[:len(norm_log_prob)-1]
        norm_log_prob = np.exp(norm_log_prob)
        return norm_log_prob

#class BHSMM_Categorical(BHSMM):
#    def __init__(self, controller):
#        BHSMM.__init__(self, controller)
#
#    def _model_init(self, dataset):
#        K = len(self._state_list)       # number of discrete states
#        D = len(self._observation_list) # dimension of the observation
#        self._hsmm = ssm.HSMM(K, D, observations='categorical')



import numpy as np
import pandas as pd
from bokeh.io import curdoc
from bokeh.models import ColumnDataSource, RadioButtonGroup, CustomJS, CheckboxGroup, Select, DataRange, DataRange1d, \
    FuncTickFormatter, LogTickFormatter, Panel, Tabs

from pyadlml.dataset import ACTIVITY
from  pyadlml.dataset.stats.activities import activities_duration_dist, activity_duration,\
    activities_transitions, activities_count, activity_duration, activities_dist
from pyadlml.dataset._core.activities import add_other_activity
from pyadlml.dataset.plot.util import func_formatter_seconds2time_log, ridgeline, \
    func_formatter_seconds2time, heatmap, annotate_heatmap, heatmap_square, savefig, \
    _num_bars_2_figsize, _num_boxes_2_figsize, \
    _num_items_2_heatmap_square_figsize, _num_items_2_ridge_figsize,\
    _num_items_2_ridge_ylimit
from pyadlml.util import get_sequential_color, get_secondary_color, get_primary_color, get_diverging_color
from bokeh.plotting import figure, show

def hist_counts(df_acts=None, lst_acts=None, df_ac=None):
    """
    Plot a bar chart displaying how often activities are occurring.

    Parameters
    ----------
    df_acts : pd.DataFrame, optional
        recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    lst_acts : lst of str, optional
        A list of activities that are included in the statistic. The list can be a
        subset of the recorded activities or contain activities that are not recorded.
    idle : bool, default: False
        Determines whether gaps between activities should be assigned
        the activity *idle* or be ignored.
    y_scale : {"log", "linear"}, default: linear
        The axis scale type to apply.

    Examples
    --------
    >>> from pyadlml.plot import plot_activity_bar_count
    >>> plot_activity_bar_count(data.df_activities, idle=True)

    .. image:: ../_static/images/plots/act_bar_cnt.png
       :height: 300px
       :width: 500 px
       :scale: 90 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing
    """
    assert not (df_acts is None and df_ac is None)

    title ='Activity occurrences'
    col_label = 'occurrence'
    x_label = 'counts'

    # create statistics if the don't exists
    if df_ac is None:
        df_acts = df_acts.copy()
        df_acts = add_other_activity(df_acts)
        df = activities_count(df_acts, lst_acts=lst_acts)
    else:
        df = df_ac

    # prepare dataframe for plotting
    df.reset_index(level=0, inplace=True)
    df = df.sort_values(by=[col_label], axis=0)

    # create a new plot with a title and axis labels
    from bokeh.io import output_file, show
    from bokeh.models import ColumnDataSource
    from bokeh.plotting import figure
    from bokeh.models.tools import HoverTool
    from bokeh.layouts import layout

    # create data column objects
    activities = df[ACTIVITY].values
    counts = df[col_label].values

    idle_idx = np.where(activities == 'idle')[0]
    activities_w_idle = np.delete(activities.copy(), idle_idx, 0)
    counts_w_idle = np.delete(counts.copy(), idle_idx, 0)

    source = ColumnDataSource(data=dict(y=activities, x=counts))
    saved_source = ColumnDataSource(data=dict(
        activities=activities,
        counts=counts))
    saved_source_w_idle = ColumnDataSource(data=dict(
        counts=counts_w_idle,
        activities=activities_w_idle,
    ))
    upper_next_10 = np.ceil(max(counts)/(10**(len(str(max(counts))))))*10**(len(str(max(counts))))# round to the highest 10th number of the maximum

    # create linear plot
    p1 = figure(title=title, y_range=activities, x_axis_label=x_label,
                x_range=[0, np.ceil(max(counts))],
               sizing_mode='stretch_width', plot_height=400,
               tools=[HoverTool()],
               tooltips="Activity @y occurred @x times.")
    p1.hbar(y='y', left=0, right='x', height=0.9, source=source,
           legend_field="y", color=get_primary_color(),
           line_color='white')

    p1.ygrid.grid_line_color = None
    p1.legend.orientation = "vertical"
    p1.legend.location = "bottom_right"
    p1.legend.click_policy = 'mute'

    # create log plot
    p2 = figure(title=title, y_range=activities, x_axis_label=x_label, x_axis_type='log',
                x_range=[0.1, upper_next_10],
               sizing_mode='stretch_width', plot_height=400,
               tools=[HoverTool()],
               tooltips="Activity @y occurred @x times.")
    p2.hbar(y='y', left=0.1, right='x', height=0.9, source=source,
           legend_field="y", color=get_primary_color(),
           line_color='white')

    p2.ygrid.grid_line_color = None
    p2.legend.orientation = "vertical"
    p2.legend.location = "bottom_right"
    p2.legend.click_policy = 'mute'

    # create widgets
    checkbox_group = create_idle_checkbox('counts', source=source, saved_source=saved_source, saved_source_w_idle=saved_source_w_idle,
                                          p1=p1, p2=p2)
    LABELS = ["Linear", "Log"]

    radio_button_group = RadioButtonGroup(labels=LABELS, active=0)

    tab1 = Panel(child=p1, title='linear')
    tab2 = Panel(child=p2, title='log')
    tabs = Tabs(tabs=[tab1, tab2], tabs_location='below', disabled=True)
    radio_button_group.js_link('active', tabs, 'active')

    layout = layout([[tabs],[checkbox_group, radio_button_group]],
        sizing_mode='stretch_width'
    )
    show(layout)

def create_idle_checkbox(col_name, source, saved_source, saved_source_w_idle, p1, p2):
    checkbox_group = CheckboxGroup(labels=["idle"], active=[0, 1])
    callback = CustomJS(
        args=dict(source=source, saved_source=saved_source, saved_source_w_idle=saved_source_w_idle, p1=p1, p2=p2),
        code="""
        var active = cb_obj.active[0];
        var sdata = saved_source.data;
        var sdata_w_idle = saved_source_w_idle.data;
        var data = source.data;

        data['x'] = []
        data['y'] = []
    
        if (active == 0){
            console.log("set data to include idle")
            data['x'] = sdata.%s
            data['y'] = sdata.activities
        }
        else {
            console.log("set data to without idle")
            data['x'] = sdata_w_idle.%s
            data['y'] = sdata_w_idle.activities
        }
        source.data = data
        //console.log(source.data)
        source.change.emit();
        
        // the plot can only be changed after the data source changes
        if (active == 0){
            p1.y_range.factors = sdata.activities
            p2.y_range.factors = sdata.activities
        }
        else {
            p1.y_range.factors = sdata_w_idle.activities
            p2.y_range.factors = sdata_w_idle.activities
        }
        p1.change.emit();
        p2.change.emit();
        """%(col_name, col_name))
    checkbox_group.js_on_click(callback)
    return checkbox_group

def hist_counts2(df_acts=None, lst_acts=None, df_ac=None):
    """
    Plot a bar chart displaying how often activities are occurring.

    Parameters
    ----------
    df_acts : pd.DataFrame, optional
        recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    lst_acts : lst of str, optional
        A list of activities that are included in the statistic. The list can be a
        subset of the recorded activities or contain activities that are not recorded.
    idle : bool, default: False
        Determines whether gaps between activities should be assigned
        the activity *idle* or be ignored.
    y_scale : {"log", "linear"}, default: linear
        The axis scale type to apply.

    Examples
    --------
    >>> from pyadlml.plot import plot_activity_bar_count
    >>> plot_activity_bar_count(data.df_activities, idle=True)

    .. image:: ../_static/images/plots/act_bar_cnt.png
       :height: 300px
       :width: 500 px
       :scale: 90 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing 
    """
    assert not (df_acts is None and df_ac is None)

    title ='Activity occurrences'
    col_label = 'occurrence'
    x_label = 'counts'

    # create statistics if the don't exists
    if df_ac is None:
        df_acts = df_acts.copy()
        df_acts = add_other_activity(df_acts)
        df = activities_count(df_acts, lst_acts=lst_acts)
    else:
        df = df_ac
    
    # prepare dataframe for plotting
    df.reset_index(level=0, inplace=True)
    df = df.sort_values(by=[col_label], axis=0)

    # create a new plot with a title and axis labels
    from bokeh.io import output_file, show
    from bokeh.models import ColumnDataSource
    from bokeh.plotting import figure
    from bokeh.models.tools import HoverTool
    from bokeh.layouts import layout

    df['log'] = np.log(df[col_label])
    df['linear'] = df[col_label]

    source = ColumnDataSource(data=dict(y=[], x=[], title=[]))
    axis_map = {
        'Linear Scaling': 'linear',
        'Log Scaling': 'log',
        0: 'linear',
        1: 'log'
    }
    x_axis = RadioButtonGroup(labels=list(axis_map.keys())[:2], active=0)

    p = figure(height=600, width=700, title="", y_range=df[ACTIVITY].values,
           sizing_mode='stretch_width',
           tools=[HoverTool()],
           tooltips="Activity @y occurred @x times.")
    p.hbar(right='x', y='y', left=0, height=0.9, source=source, legend_field="y", color=get_primary_color(), line_color='white')

    def update():
        #p.y_range = DataRange1d(df[ACTIVITY].values)
        print('#'*100)
        x_name = axis_map[x_axis.active]
        if x_name == 'linear':
            p.xaxis.axis_label = 'count'
        else:
            p.xaxis.axis_label = 'log count'
            #p.xaxis.formatter = FuncTickFormatter(code="""return Math.log(tick).toFixed(2)""")
            p.xaxis.formatter = LogTickFormatter()
        p.title.text = "Test"
        print(df[x_name].values)
        print(df[ACTIVITY].values)
        source.data = dict(
            x=df[x_name].values,
            y=df[ACTIVITY].values,
        )

    x_axis.on_change('active', lambda attr, old, new: update())
    layout = layout([[p],[x_axis]],
        sizing_mode='stretch_width'
    )
    update()
    curdoc().add_root(layout)
    curdoc().title = "Test 5"
    #show(layout)


def boxplot_duration(df_acts, lst_acts=None, file_path=None):
    """
    Plot a boxplot for activity durations.

    Parameters
    ----------
    df_acts : pd.DataFrame, optional
        recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    lst_acts : lst of str, optional
        A list of activities that are included in the statistic. The list can be a
        subset of the recorded activities or contain activities that are not recorded.
    figsize : (float, float), default: None
        width, height in inches. If not provided, the figsize is inferred by automatically.
    y_scale : {"log", "linear"}, default: None
        The axis scale type to apply.
    idle : bool, default: False
        Determines whether gaps between activities should be assigned
        the activity *idle* or be ignored.
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import plot_activity_bp_duration
    >>> plot_activity_bp_duration(data.df_activities)

    .. image:: ../_static/images/plots/act_bp.png
       :height: 300px
       :width: 500 px
       :scale: 90 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing
    """
    title = 'Activity durations'
    xlabel = 'seconds'

    df_acts = add_other_activity(df_acts)

    df = activities_duration_dist(df_acts, lst_acts=lst_acts)

    # select data for each device
    activities = df[ACTIVITY].unique()
    df['seconds'] = df['minutes']*60     

    num_act = len(activities)

    dat = []
    for activity in activities:
        df_activity = df[df[ACTIVITY] == activity]
        #tmp = np.log(df_device['td'].dt.total_seconds())
        dat.append(df_activity['seconds']) 
    
    # plot boxsplot
    fig, ax = plt.subplots(figsize=figsize)
    ax.boxplot(dat, vert=False)
    ax.set_title(title)
    ax.set_yticklabels(activities, ha='right')
    ax.set_xlabel(xlabel)
    ax.set_xscale('log')

    # create secondary axis with time format 1s, 1m, 1d
    ax_top = ax.secondary_xaxis('top', functions=(lambda x: x, lambda x: x))
    #ax_top.set_xlabel('time')
    ax_top.xaxis.set_major_formatter(
        ticker.FuncFormatter(func_formatter_seconds2time))

    if file_path is not None:
        savefig(fig, file_path)
        return 
    else:
        return fig

def hist_cum_duration(df_acts=None, lst_acts=None, df_dur=None, file_path=None):
    """
    Plots the cumulative duration for each activity in a bar plot.

    Parameters
    ----------
    df_acts : pd.DataFrame, optional
        recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    lst_acts : lst of str, optional
        A list of activities that are included in the statistic. The list can be a
        subset of the recorded activities or contain activities that are not recorded.
    y_scale : {"log", "linear"}, default: None
        The axis scale type to apply.
    idle : bool, default: False
        Determines whether gaps between activities should be assigned
        the activity *idle* or be ignored.
    figsize : (float, float), default: None
        width, height in inches. If not provided, the figsize is inferred by automatically.
    color : str, optional
        sets the color of the plot. When not set, the primary theming color is used.
        Learn more about theming in the :ref:`user guide <theming>`
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import plot_activity_bar_duration
    >>> plot_activity_bar_duration(data.df_activities)

    .. image:: ../_static/images/plots/act_bar_dur.png
       :height: 300px
       :width: 500 px
       :scale: 90 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing
    """
    assert not (df_acts is None and df_dur is None)

    title = 'Cummulative activity durations'
    x_label = 'seconds'
    freq = 'seconds'

    if df_dur is None:
        df_acts = add_other_activity(df_acts.copy())
        df = activity_duration(df_acts, lst_acts=lst_acts, time_unit=freq)
    else:
        df = df_dur
    df = df.sort_values(by=[freq], axis=0)

    # create a new plot with a title and axis labels
    from bokeh.io import output_file, show
    from bokeh.models import ColumnDataSource
    from bokeh.plotting import figure
    from bokeh.models.tools import HoverTool
    from bokeh.layouts import layout

    col_label = freq

    # create data column objects
    activities = df[ACTIVITY].values
    counts = df[col_label].values

    idle_idx = np.where(activities == 'idle')[0]
    activities_w_idle = np.delete(activities.copy(), idle_idx, 0)
    counts_w_idle = np.delete(counts.copy(), idle_idx, 0)

    source = ColumnDataSource(data=dict(y=activities, x=counts))
    saved_source = ColumnDataSource(data=dict(
        activities=activities,
        counts=counts))
    saved_source_w_idle = ColumnDataSource(data=dict(
        counts=counts_w_idle,
        activities=activities_w_idle,
    ))
    upper_next_10 = np.ceil(max(counts)/(10**(len(str(max(counts))))))*10**(len(str(max(counts))))# round to the highest 10th number of the maximum

    # create linear plot
    p1 = figure(title=title, y_range=activities, x_axis_label=x_label,
                x_range=[0, np.ceil(max(counts))],
               sizing_mode='stretch_width', plot_height=400,
               tools=[HoverTool()],
               tooltips="Activity @y occurred @x seconds")
    p1.hbar(y='y', left=0, right='x', height=0.9, source=source,
           legend_field="y", color=get_primary_color(),
           line_color='white')

    p1.ygrid.grid_line_color = None
    p1.legend.orientation = "vertical"
    p1.legend.location = "bottom_right"
    p1.legend.click_policy = 'mute'

    # create log plot
    p2 = figure(title=title, y_range=activities, x_axis_label=x_label, x_axis_type='log',
                x_range=[0.1, upper_next_10],
               sizing_mode='stretch_width', plot_height=400,
               tools=[HoverTool()],
               tooltips="Activity @y occurred @x times.")
    p2.hbar(y='y', left=0.1, right='x', height=0.9, source=source,
           legend_field="y", color=get_primary_color(),
           line_color='white')

    p2.ygrid.grid_line_color = None
    p2.legend.orientation = "vertical"
    p2.legend.location = "bottom_right"
    p2.legend.click_policy = 'mute'

    # create widgets
    checkbox_group = create_idle_checkbox('counts', source=source, saved_source=saved_source, saved_source_w_idle=saved_source_w_idle,
                                          p1=p1, p2=p2)
    LABELS = ["Linear", "Log"]

    radio_button_group = RadioButtonGroup(labels=LABELS, active=0)

    tab1 = Panel(child=p1, title='linear')
    tab2 = Panel(child=p2, title='log')
    tabs = Tabs(tabs=[tab1, tab2], tabs_location='below', disabled=True)
    radio_button_group.js_link('active', tabs, 'active')

    layout = layout([[tabs],[checkbox_group, radio_button_group]],
        sizing_mode='stretch_width'
    )
    show(layout)

def heatmap_transitions(df_acts=None, lst_acts=None, df_trans=None, z_scale="linear",
                        figsize=None, idle=False, numbers=True, grid=True,
                        cmap=None, file_path=None):
    """
    Parameters
    ----------
    df_acts : pd.DataFrame, optional
        recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    lst_acts : lst of str, optional
        A list of activities that are included in the statistic. The list can be a
        subset of the recorded activities or contain activities that are not recorded.
    df_trans : pd.DataFrame
        A precomputed transition table. If the *df_trans* parameter is given, parameters
        *df_acts* and *lst_acts* are ignored. The transition table can be computed
        in :ref:`stats <stats_acts_trans>`.
    figsize : (float, float), default: None
        width, height in inches. If not provided, the figsize is inferred by automatically.
    z_scale : {"log", "linear"}, default: None
        The axis scale type to apply.
    numbers : bool, default: True
        Whether to display numbers inside the heatmaps fields or not.
    idle : bool, default: False
        Determines whether gaps between activities should be assigned
        the activity *idle* or be ignored.
    cmap : str or Colormap, optional
        The Colormap instance or registered colormap name used to map scalar
        data to colors. This parameter is ignored for RGB(A) data.
        Defaults 'viridis'.
    grid : bool, default: True
        determines whether to display a white grid, seperating the fields or not.
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import plot_activity_hm_transition
    >>> plot_activity_hm_transition(data.df_activities)

    .. image:: ../_static/images/plots/act_hm_trans.png
       :height: 300px
       :width: 500 px
       :scale: 90 %
       :alt: alternate text
       :align: center


    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing.
    """
    assert z_scale in ['linear', 'log'], 'z-scale has to be either of type None or log'
    assert not (df_acts is None and df_trans is None)

    title = 'Activity transitions'
    z_label = 'count'

    if df_trans is None:
        df_acts = add_other_activity(df_acts) if idle else df_acts
        df = activities_transitions(df_acts, lst_acts=lst_acts)
    else:
        df = df_trans

    # get the list of cross tabulations per t_window
    act_lst = list(df.columns)

    num_act = len(act_lst)
    figsize = (_num_items_2_heatmap_square_figsize(num_act) if figsize is None else figsize)
    cmap = (get_sequential_color() if cmap is None else cmap)

    x_labels = act_lst
    y_labels = act_lst
    values = df.values
    

    log = True if z_scale == 'log' else False
    valfmt = '{x:.0f}'
        
     # begin plotting
    fig, ax = plt.subplots(figsize=figsize)
    im, cbar = heatmap_square(values, y_labels, x_labels, log=log, cmap=cmap, ax=ax, cbarlabel=z_label, grid=grid)
    if numbers:
        texts = annotate_heatmap(im, textcolors=("white", "black"),log=log, valfmt=valfmt)
    ax.set_title(title)
    
    if file_path is not None:
        savefig(fig, file_path)
        return 
    else:
        return fig

def ridge_line(df_acts=None, lst_acts=None, df_act_dist=None, idle=False,
               n=1000, ylim_upper=None, color=None, figsize=None, file_path=None):
    """
    Plots the activity density distribution over one day.

    Parameters
    ----------
    df_acts : pd.DataFrame, optional
        recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    lst_acts : lst of str, optional
        A list of activities that are included in the statistic. The list can be a
        subset of the recorded activities or contain activities that are not recorded.
    df_act_dist : pd.DataFrame, optional
        A precomputed activity density distribution. If the *df_trans* parameter is given, parameters
        *df_acts* and *lst_acts* are ignored. The transition table can be computed
        in :ref:`stats <stats_acts_trans>`.
    n : int, default=1000
        The number of monte-carlo samples to draw.
    ylim_upper: float, optional
        The offset from the top of the plot to the first ridge_line. Set this if
        the automatically determined value is not satisfying.
    figsize : (float, float), default: None
        width, height in inches. If not provided, the figsize is inferred by automatically.
    color : str, optional
        sets the color of the plot. When not set, the primary theming color is used.
        Learn more about theming in the :ref:`user guide <theming>`
    idle : bool, default: False
        Determines whether gaps between activities should be assigned
        the activity *idle* or be ignored.
    file_path : str, optional
        If set, saves the plot under the given file path and return *None* instead
        of returning the figure.

    Examples
    --------
    >>> from pyadlml.plot import plot_activity_rl_daily_density
    >>> plot_activity_rl_daily_density(data.df_activities)

    .. image:: ../_static/images/plots/act_ridge_line.png
       :height: 300px
       :width: 500 px
       :scale: 90 %
       :alt: alternate text
       :align: center

    Returns
    -------
    res : fig or None
        Either a figure if file_path is not specified or nothing.
    """
    assert not (df_acts is None and df_act_dist is None)

    title = 'Activity distribution over one day'
    xlabel = 'day'
    color = (get_primary_color() if color is None else color)


    if df_act_dist is None:
        if idle:
            df_acts = add_other_activity(df_acts)
        df = activities_dist(df_acts.copy(), lst_acts=lst_acts, n=n)
        if df.empty:
            raise ValueError("no activity was recorded and no activity list was given.")
    else:
        df = df_act_dist

    def date_2_second(date):
        """ maps time onto seconds of a day 
        Parameters
        ----------
        date : np.datetime64
            all the dates are on the day 1990-01-01

        Returns
        -------

        """
        if pd.isnull(date):
            return -1
        val = (date - np.datetime64('1990-01-01')) / np.timedelta64(1, 's')
        total_seconds = 60*60*24
        assert val <= total_seconds and val >= 0
        return int(val)

    
    df = df.apply(np.vectorize(date_2_second))
    # sort every columns values ascending
    for col in df.columns:
        df[col] = df[col].sort_values()

    grouped = [(col, df[col].values) for col in df.columns]
    acts, data = zip(*grouped)
    num_act = len(list(acts))

    # infer visual properties
    figsize = (_num_items_2_ridge_figsize(num_act) if figsize is None else figsize)
    ylim_upper = (_num_items_2_ridge_ylimit(num_act) if ylim_upper is None else ylim_upper)

    # plot the ridgeline
    fig, ax = plt.subplots(figsize=figsize)
    ridgeline(data, labels=acts, overlap=.85, fill=color, n_points=100, dist_scale=0.13)
    plt.title(title)

    plt.gca().spines['left'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)
    plt.gca().spines['top'].set_visible(False)
    plt.ylim((0, ylim_upper))
    plt.xlabel(xlabel)
    
    # set xaxis labels
    def func(x,p):
        #x = x + 0.5
        #if x == 0.0 or str(x)[-1:] == '5':
        #    return ''
        #else:
        if True:
            if np.ceil(x/k) < 10:
                return '0{}:00'.format(int(x/k)+1)
            else:
                return '{}:00'.format(int(x/k)+1)
    a = 0
    b = 60*60*24
    k = (b-a)/24
    
    plt.xlim((a,b))
    tcks_pos = np.arange(0,23)*k + (-0.5 + k)
    
    x_locator = ticker.FixedLocator(tcks_pos)
    ax.xaxis.set_major_formatter(ticker.FuncFormatter(func))
    ax.xaxis.set_major_locator(x_locator)
    fig.autofmt_xdate(rotation=45)
    
    plt.grid(zorder=0)

    if file_path is not None:
        savefig(fig, file_path)
        return 
    else:
        return fig


from pyadlml.dataset.stats.acts_and_devs import (
    contingency_table_events as contingency_events,
    contingency_table_states as contingency_states,
    cross_correlogram as cross_correlogram
)

from pyadlml.dataset.stats.activities import (
    activities_dist as activity_dist,
    activities_count as activity_count,
    activities_transitions as activity_transition,
    activity_duration as activity_duration,
    activities_duration_dist as activity_duration_dist
)

from pyadlml.dataset.stats.devices import (
    state_fractions as device_state_fractions,
    state_cross_correlation as device_state_cross_correlation,
    state_times as device_time_in_states,
    inter_event_intervals as device_inter_event_intervals,
    events_one_day as device_event_density,
    event_cross_correlogram as device_event_cross_correlogram,
    event_count as device_event_count,
)


import functools
import inspect
import joblib
from copy import copy

import pandas as pd
import numpy as np

from pyadlml.constants import DATASET_STRINGS, TIME, END_TIME, START_TIME, DEVICE, VALUE, BOOL, \
    CAT, ACTIVITY
from pandas.api.types import infer_dtype
from pyadlml.dataset._core.activities import ActivityDict, _is_activity_overlapping, \
    correct_activity_overlap

"""
    includes generic methods for manpulating dataframes
"""

def print_df(df):
    with pd.option_context('display.max_rows', None, 'display.max_columns', None):  # more options can be specified also
        print(df)

def unitsfromdaystart(ts, unit='s'):
    """ Computes units passed from the start of the day until the timestamp
    """
    ts = pd.Timestamp(ts) if not isinstance(ts, pd.Timestamp) else ts
    seconds = ts.hour*3600 + ts.minute*60 + ts.second
    if unit == 's':
        return int(seconds)
    elif unit == 'm':
        return int(seconds/60)
    elif unit == 'h':
        return int(seconds/3600)
    else:
        raise ValueError


def timestr_2_timedeltas(t_strs):
    """
        gets either a string or a list of strings to convert to a list of 
        timedeltas
    """
    if isinstance(t_strs, list):
        return [timestr_2_timedelta(t_str) for t_str in t_strs]
    else:
        return [timestr_2_timedelta(t_strs)]


def timestr_2_timedelta(t_str):
    """
        t_str (string)
        of form 30s, 30m
    """
    ttype = t_str[-1:]
    val = int(t_str[:-1])

    assert ttype in ['h','m','s']
    assert (val > 0 and val <=12 and ttype == 'h')\
        or (val > 0 and val <= 60 and ttype == 'm')\
        or (val > 0 and val <= 60 and ttype == 's')
    import datetime as dt
    if ttype == 's':
        return pd.Timedelta(seconds=val)
    if ttype == 'm':
        return pd.Timedelta(seconds=val*60)
    if ttype == 'h':
        return pd.Timedelta(seconds=val*3600)


def time2int(ts, t_res='30m'):
    """
    rounds to the next lower min bin or hour bin
    """
    assert t_res[-1:] in ['h','m']
    val = int(t_res[:-1])

    assert (val > 0 and val <=12 and t_res[-1:] == 'h')\
        or (val > 0 and val <= 60 and t_res[-1:] == 'm')

    import datetime as dt
    zero = dt.time()

    if t_res[-1:] == 'h':
        hs = val
        h_bin = int(ts.hour/hs)*hs
        return dt.time(hour=h_bin)

    elif t_res[-1:] == 'm':
        ms = val
        m_bin = int(ts.minute/ms)*ms
        return dt.time(hour=ts.hour, minute=m_bin)
    else:
        raise ValueError



def fill_nans_ny_inverting_first_occurence(df):
    """
    fills up true or false values
    :param df:
                Name           0      1         10
        time                                     
        2008-02-25 00:20:14    NaN    NaN   ...    True
        2008-02-25 00:22:57    NaN    NaN   ...    False
        2008-02-25 09:33:41    NaN    True  ...    True
        2008-02-25 09:33:42    False    NaN   ...    False
    :return:
                Name           0      1        10   
        Time                                       
        2008-02-25 00:20:14    True   False ... True 
        2008-02-25 00:22:57    True   False ... False 
        2008-02-25 09:33:41    True   True  ... True 
        2008-02-25 09:33:42    False   NaN  ... False 
    """
    for col_label in df.columns:
        col = df[col_label]

        # get timestamp of first valid index and replace previous Nans 
        #   by opposite
        ts = col.first_valid_index()
        idx = df.index.get_loc(ts)
        col.iloc[0:idx] = not col[ts]
    return df


def categorical_2_binary(df_devices, cat_list):
    """
    Transforms all categorical devices within the device dataframe into
    binary by creating a new device 'cat:cat_value' that reports a 1
    when the category is activated and 0 when it ends

    Parameters
    ----------
    df_devices : pd.DataFrame
        A device dataframe
    cat_list : lst
        Categorical devices from the dataframe

    Returns
    -------
    res : pd.DataFrame
        A device dataframe with newly created devices

    """
    mask_cat = df_devices[DEVICE].isin(cat_list)
    df_cat = df_devices[mask_cat].copy()
    df_cat.loc[:, 'new_device'] = df_cat[DEVICE] + ':' + df_cat[VALUE]
    df_cat.loc[:, VALUE] = True
    for dev in df_cat[DEVICE].unique():
        df = df_cat[df_cat[DEVICE] == dev]
        df_new = df.copy()
        df_new.loc[:, DEVICE] = df['new_device'].shift(1)
        df_new = df_new[1:]
        df_new.loc[:, VALUE] = False
        df_new.loc[:, TIME] += pd.Timedelta('1ns')
        df_cat = pd.concat([df_cat, df_new])

    mask = df_cat[DEVICE].isin(cat_list)
    df_cat.loc[mask, DEVICE] = df_cat.loc[mask, 'new_device']
    df_cat = df_cat.drop(columns='new_device')

    df_devices = pd.concat([df_devices[~mask_cat], df_cat])\
                    .sort_values(by=TIME)\
                    .reset_index(drop=True)
    return df_devices

def infer_dtypes(df_devices):
    """ Infers automatically the datatypes for each device of a device dataframe
    and returns a dictionary containing data types mapped to device names

    Parameters
    ----------
    df_devices : pd.DataFrame
        A device dataframe

    Returns
    -------
    res : dict
        Dictionary with keys: {'categorical' : [...], 'boolean': [...] , 'numerical' : [...]}

    """
    dev_cat = []
    dev_bool = []
    dev_num = []

    dev_lst = df_devices[DEVICE].unique()
    for dev in dev_lst:
        vals = df_devices[df_devices[DEVICE] == dev][VALUE]
        inf = infer_dtype(vals, skipna=True)
        if inf == 'string' or inf == 'object' or inf == 'mixed':
            try:
                pd.to_numeric(vals.dropna().unique())
                dev_num.append(dev)
            except:
                dev_cat.append(dev)
        elif inf == 'boolean':
            dev_bool.append(dev)
        elif inf == 'floating':
            dev_num.append(dev)
        else:
            raise ValueError('could not infer correct dtype for device {}'.format(dev))

    return {'categorical': dev_cat, 'boolean': dev_bool, 'numerical': dev_num}


def select_timespan(df_devs=None, df_acts=None, start_time=None, end_time=None, clip_activities=False):
    """ Selects a subset of a device and an activity dataframe based on a time span given
    Parameters
    ----------
    df_devices : pd.DataFrame or None
        A device dataframe
    df_activities : pd.DataFrame, dict containing dataframes or None
        An activity dataframe
    start_time : str or None
        The start time from
    end_time : str or None
    
    clip_activities : bool, default=False
        If set then the activities are clipped to the start and end time

    Returns
    -------
    df_d, df_a : the two subsets
    """
    # Cast to pandas timestamp
    start_time = str_to_timestamp(start_time) if isinstance(start_time, str) else start_time
    end_time = str_to_timestamp(end_time) if isinstance(end_time, str) else end_time

    if df_devs is not None:
        if start_time is not None:
            dev_start_sel = (df_devs[TIME] >= start_time)
        else:
            dev_start_sel = np.full(len(df_devs), True)
        if end_time is not None:
            dev_end_sel = (df_devs[TIME] < end_time)
        else:
            dev_end_sel = np.full(len(df_devs), True)
        df_devs = df_devs[dev_start_sel & dev_end_sel]

    if df_acts is not None:
        df_acts_inst_type = type(df_acts)
        df_acts = ActivityDict.wrap(df_acts)

        for k in df_acts.keys():
            df_activity = df_acts[k]
            if start_time is not None:
                act_start_sel = ~(df_activity[END_TIME] < start_time)
            else:
                act_start_sel = np.full(len(df_activity), True)
            if end_time is not None:
                act_end_sel = ~(df_activity[START_TIME] > end_time)
            else:
                act_end_sel = np.full(len(df_activity), True)

            df_activity = df_activity[act_start_sel & act_end_sel]

            # clip activities if they extend in regions that are not in the timespan
            if not df_activity.empty and start_time is not None \
                    and (start_time < df_activity[END_TIME].iat[0] and start_time > df_activity[START_TIME].iat[0]):
                df_activity.at[df_activity.index[0], START_TIME] = start_time
            if not df_activity.empty and clip_activities and end_time is not None \
            and end_time < df_activity[END_TIME].iat[-1]:
                df_activity.at[df_activity.index[-1], END_TIME] = end_time
                
            df_acts[k] = df_activity

        df_acts = df_acts.unwrap(df_acts_inst_type) 

    if df_acts is None:
        return df_devs
    elif df_devs is None:
        return df_acts
    else:
        return df_devs, df_acts


def get_last_states(df_devs: pd.DataFrame, left_bound=-1) -> dict:
    """ Creates a dictionary where every device maps to its last known value.

    Parameters
    ----------
    df_devs : pd.DataFrame
        TODO add description 
    left_bound : int
        The index of the first event from which the states are determined directed
        at the past 

    Returns
    -------
    res : dict
        A device mapping to initial values
    """
    left_bound = len(df_devs) if left_bound == -1 else left_bound
    return df_devs.iloc[:left_bound, :]\
            .sort_values(by=TIME)\
            .groupby(DEVICE)\
            .last()[VALUE]\
            .to_dict()

def get_first_states(df_devs: pd.DataFrame) -> dict:
    """ Creates a dictionary where every device maps to its first known value.

    Parameters
    ----------
    df_devs : pd.DataFrame
        TODO add description 

    Returns
    -------
    res : dict
        A device mapping to initial values
    """
    return df_devs.copy()\
        .sort_values(by=TIME)\
        .groupby(DEVICE)\
        .first()[VALUE]\
        .to_dict()



def str_to_timestamp(val):
    """ Converts a datetime string to a panda Timestamp.
    the day-first format is used.

    Parameters
    ----------
    val : list or string

    Returns
    -------
    pd.Timestamp

    """
    return pd.to_datetime(val, dayfirst=True)


def df_difference(df1: pd.DataFrame, df2: pd.DataFrame, which=None):
    """Find rows which are different between two DataFrames.

    Parameters
    ----------
    df1 : pd.DataFrame
        TODO
    df2 : pd.DataFrame
        TODO
    which : None or str, default=None
        TODO


    Returns
    -------

    """
    from pyadlml.dataset._core.devices import is_device_df
    from pyadlml.dataset._core.activities import is_activity_df

    comparison_df = df1.copy().merge(
        df2.copy(),
        indicator=True,
        how='outer'
    )
    if which is None:
        diff_df = comparison_df[comparison_df['_merge'] != 'both']
    else:
        diff_df = comparison_df[comparison_df['_merge'] == which]

    if is_activity_df(df1):
        return diff_df[[START_TIME, END_TIME, ACTIVITY]]
    elif is_device_df(df1):
        return diff_df[[TIME, DEVICE, VALUE]]
    else:
        return diff_df




def event_times(df_devices, start_time=None, end_time=None):
    """
    Parameters
    ----------
    time_array : nd.array or pd.Series or pd.DataFrame
    start_time : pd.Timestamp

    end_time : pd.Timestamp


    Returns
    -------
    res : nd.array
        Array of transformed timestamps
    start_time : pd.Timestamp

    end_time : pd.Timestamp
    """
    if isinstance(time_array, pd.DataFrame):
        time_array = time_array[TIME].values
    if isinstance(time_array, pd.Series):
        time_array = time_array.values

    # get start and end_time
    if start_time is None:
        start_time = time_array[0]
    if end_time is None:
        end_time = time_array[-1]

    # map to values between [0,1]
    res = (time_array - start_time)/(end_time - start_time)

    return res, start_time, end_time


def num_to_timestamp(val, start_time, end_time):
    """Converts value [0,1] into timestamp between start_time and end_time"""
    return start_time + val*(end_time - start_time)

def timestamp_to_num(ts, start_time, end_time):
    """Converts timestamp between start_time and end_time into value in [0,1]"""
    return float((ts - start_time)/(end_time - start_time))

def get_sorted_index(df: pd.DataFrame, rule='alphabetical', area: pd.DataFrame = None) -> np.ndarray:
    """ Returns a new dataframes index that is sorted after a specific rule

    Parameters
    ----------
    df : pd.DataFrame
        Has to contain the column 'activity' or 'device'

    rule : one of {list, ndarray, 'alphabetical', 'area', str}
        The rule on how to order the array.

        - alphabetical : the

    Returns
    -------
    new_order : ndarray
        An array with the new ordered indices for the dataframe
    """
    rule_is_iter = (isinstance(rule, list) or isinstance(rule, np.ndarray))

    # If df is a list the enumeration for that list is returned
    if (isinstance(df, list) or isinstance(df, np.ndarray)) and rule_is_iter:
        mapping = {v: k for k, v in enumerate(rule)}
        return np.vectorize(mapping.get)(df)

    df = df.copy()
    df['order'] = np.arange(len(df))
    new_order = []

    if rule_is_iter:
        mapping = {v: k for k, v in enumerate(rule)}
        order_changed = False
        try:
            df['order'] = df[ACTIVITY].map(mapping)
            order_changed = True
        except:
            pass
        try:
            df['order'] = df[DEVICE].map(mapping)
            order_changed = True
        except:
            pass
        assert order_changed, '"activity" or "device" was not present in dataframe'
    elif rule == 'alphabetical':
        if ACTIVITY in df.columns:
            df = df.sort_values(by=ACTIVITY)
        elif DEVICE in df.columns:
            df = df.sort_values(by=DEVICE)
        else:
            raise KeyError(f"Tried to sort alphabetical but no activity or device column was found, only {rule}")
    elif rule == 'value':
        # The case when the column other than ACTIVITY or order should be used
        cs = df.columns
        col = [i for i in cs if i not in [ACTIVITY, DEVICE, 'order']]
        assert len(col) == 1, 'When value is specified there must be only one remaining column.'
        if ACTIVITY in df.columns:
            df = df.sort_values(by=col[0])
        elif DEVICE in df.columns:
            df = df.sort_values(by=col[0])
        else:
            raise KeyError(f"Tried to sort alphabetical but no activity or device column was found, only {rule}")
    elif rule == 'areas':
        raise NotImplementedError
    elif isinstance(rule, str):
        assert rule in df.columns, f'The rule {rule} was not in the dataframes columns.'
        df = df.sort_values(by=rule)
    else:
        raise KeyError()
    new_order = df['order'].values

    return new_order


def extract_kwargs(func):
    """ Gets a function end sets all args to kwargs and all default values from
        the signature.
    """
    @functools.wraps(func)
    def wrapper_extract(*args, **kwargs):
        sign = inspect.signature(func).parameters

        # assign the args to the respective kwargs
        for arg_name, arg in zip(sign, args):
            kwargs[arg_name] = arg

        # set the rest of the kwargs to the default values
        for param in sign.values():
            if param.name not in kwargs.keys():
                kwargs[param.name] = param.default

        return func(**kwargs)
    return wrapper_extract


def check_order(func):
    @extract_kwargs
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        o = kwargs['order']
        assert o in ['alphabetical', 'occurence'] or isinstance(o, list) \
               or isinstance(o, np.ndarray), f'Sort rule is either "alphabetical", a custom string or a iterable. Found {o}'
        tmp = func(*args, **kwargs)
        return tmp
    return wrapper

def check_scale(func):
    @extract_kwargs
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        assert kwargs['scale'] in ['linear', 'log']
        return func(*args, **kwargs)
    return wrapper


def device_order_by(df_devs: pd.DataFrame, rule='alphabetical', dev2area=None) -> list:
    """Create a list of devices ordered by some rule

    Returns
    -------
    list of strings
    """
    if isinstance(rule, list) or isinstance(rule, np.ndarray):
        return rule

    allowed_combis =['alphabetical', 'count', 'area', 'area+alphabetical', 'area+count'] 
    assert rule in allowed_combis, f'rule{rule} not in assertion.'

    if rule == 'alphabetical':
        dev_order = df_devs[DEVICE].unique().tolist()
        dev_order.sort() # CAVE returns None, do not change ^^
    elif rule == 'count':
        from pyadlml.dataset.stats.devices import device_order_by_count
        dev_order = device_order_by_count(df_devs)
    elif 'area' in rule:
        assert dev2area is not None, 'When selecting area as sorting measure, an area mapping must be given as parameter.'
        split = rule.split('+')
        dev2area['area'] = dev2area['area'].fillna('-na')
        if len(split) == 1 or split[1] == 'alphabetical' :
            dev_order = dev2area.groupby('area')\
                                .apply(lambda x: x.sort_values(by=DEVICE))\
                                .reset_index(drop=True)[DEVICE].to_list()
        elif split[1] == 'count':
            from pyadlml.dataset.stats.devices import device_order_by_count
            ordered_by_count = device_order_by_count(df_devs)
            dev_order = pd.DataFrame(zip(range(len(ordered_by_count), ordered_by_count)), columns=['count', DEVICE])
            dev_order = dev_order.set_index(DEVICE)
            tmp = dev2area.set_index(DEVICE)
            tmp['count'] = dev_order['count']
            dev_order = dev2area.groupby('area')\
                                .apply(lambda x: x.sort_values(by='count'))\
                                .reset_index(drop=True)[DEVICE].to_list()

    return dev_order

def activity_order_by(dct_acts, rule: str='alphabetical') -> list:
    """ Return a list with ordered activities

    Parameters
    ----------
    dct_acts: pd.DataFrame or ActivityDict
    rule: str
        The rule one of:
            - alphabetical
            - duration
            - count
            - area
    Returns
    -------
    list of str
        The ordered activity names
    """
    is_iterable = isinstance(rule, list) or isinstance(rule, np.ndarray)
    if is_iterable:
        # TODO refactor, negate: any not int and not dct is none than below 
        if all(isinstance(item, int) for item in rule) or dct_acts is None:
            return rule

        # Filter out activities that are not in df_acts
        rule = copy(list(rule))
        for act_not_in_df in set(rule) - set(dct_acts[ACTIVITY].unique()):
            rule.remove(act_not_in_df)
        return rule

    assert rule in ['alphabetical', 'duration', 'count', 'area'], f'rule{rule} not in assertion.'

    if isinstance(dct_acts, pd.DataFrame):
        dct_acts = ActivityDict.wrap(dct_acts)

    if rule == 'alphabetical':
        act_order = dct_acts.get_activity_union()
        act_order.sort()
    elif rule == 'duration':
        from pyadlml.dataset.stats.activities import activity_order_by_duration
        act_order = activity_order_by_duration(dct_acts)
    elif rule == 'count':
        from pyadlml.dataset.stats.activities import activity_order_by_count
        act_order = activity_order_by_count(dct_acts)
    else:
        raise NotImplementedError('Area has still to be implemented.')

    return act_order


def fetch_by_name(dataset: str, identifier=None, **kwargs) -> dict:
    """ Fetches a dataset based on a given string. 

    Parameters
    ----------
    dataset: str


    Returns
    -------
    dict
    """
    error_msg = f'No suitable option specified: {dataset}.\nAvailable are {DATASET_STRINGS}'
    assert dataset in DATASET_STRINGS + ['joblib'], error_msg

    from pyadlml.dataset import fetch_amsterdam, \
                                fetch_mitlab, fetch_aras, fetch_kasteren_2010, \
                                fetch_casas_aruba, fetch_tuebingen_2019, \
                                fetch_uci_adl_binary, load_act_assist
    if dataset == DATASET_STRINGS[0]:
        return fetch_casas_aruba(**kwargs)
    elif dataset == DATASET_STRINGS[1]:
        return fetch_amsterdam(**kwargs)
    elif dataset == DATASET_STRINGS[2]:
        return fetch_mitlab(subject='subject1', **kwargs)
    elif dataset == DATASET_STRINGS[3]:
        return fetch_mitlab(subject='subject2', **kwargs)
    elif dataset == DATASET_STRINGS[4]:
        return fetch_aras(**kwargs)
    elif dataset == DATASET_STRINGS[5]:
        return fetch_kasteren_2010(house='A', **kwargs)
    elif dataset == DATASET_STRINGS[6]:
        return fetch_kasteren_2010(house='B', **kwargs)
    elif dataset == DATASET_STRINGS[7]:
        return fetch_kasteren_2010(house='C', **kwargs)
    elif dataset == DATASET_STRINGS[8]:
        return fetch_tuebingen_2019(**kwargs)
    elif dataset == DATASET_STRINGS[9]:
        return fetch_uci_adl_binary(subject='OrdonezA', **kwargs)
    elif dataset == DATASET_STRINGS[10]:
        return fetch_uci_adl_binary(subject='OrdonezB', **kwargs)
    elif dataset == DATASET_STRINGS[11]:
        return load_act_assist(identifier)
    elif dataset == 'joblib':
        return joblib.load(identifier)


def get_dev_row_where(df, time, dev, state):
    raise ValueError('deprecated: use get_index_matchingrows instead')
    time = pd.Timestamp(time)
    mask = (df[DEVICE] == dev) \
           & (df[TIME] == time) \
           & (df[VALUE] == state)
    df = df.reset_index().set_index('index')
    return df[mask.values].copy()

def get_dev_rows_where(df_devs, rows):
    raise ValueError('deprecated: use get_index_matchingrows instead')
    res = [get_dev_row_where(df_devs, r[0], r[1], r[2]) for r in rows]
    return pd.concat(res)

def append_devices(df_devs, rows):
    raise ValueError('deprecated: use get_index_matchingrows instead')
    df = pd.DataFrame(data=rows, columns=[TIME, DEVICE, VALUE])
    df[TIME] = pd.to_datetime(df[TIME])
    return pd.concat([df_devs, df])

def remove_devices(df_devs, rows):
    raise ValueError('deprecated: use get_index_matchingrows instead')
    idx_to_drop = [get_dev_row_where(df_devs, r[0], r[1], r[2]).index[0] for r in rows]
    return df_devs.drop(index=idx_to_drop)



import numpy as np

def create_xset_onebitflip(last_observation):
    """ constructs an array with all observations, that can happen if only one
    sensor flips a bit or not
    Parameters
    ----------
    last_observation array_like (D, D)
        the last observation that was made
    Returns
    -------
    result array_like (D+1, D)
        the last row contains the log probability of nothing happening
    """
    sample_xs = last_observation
    dim = len(last_observation)
    res = np.zeros((dim+1, dim), dtype=np.int64)
    mask = np.eye(dim, dtype=np.int64)
    for i in range(dim):
        res[i] = np.logical_xor(last_observation, mask[i])
    res[dim] = last_observation
    return res



import sys
import pathlib
working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))
import unittest


from pyadlml.dataset import set_data_home, load_act_assist
from pyadlml.dataset import ACTIVITY, DEVICE, END_TIME, START_TIME, VAL, \
    TIME

SUBJECT_ONE_NAME = 'one'
SUBJECT_TWO_NAME = 'two'
SUBJECT_ADMIN_NAME = 'admin'

class TestEmptyDataset(unittest.TestCase):
    def setUp(self):
        dataset_dir = str(script_directory) + '/datasets/empty_dataset'
        self.data = load_act_assist(dataset_dir,
            subjects=[SUBJECT_ONE_NAME, SUBJECT_TWO_NAME])

    def test_activities_loaded(self):
        df_one = getattr(self.data, 'df_activities_{}'.format(SUBJECT_ONE_NAME))
        df_two = getattr(self.data, 'df_activities_{}'.format(SUBJECT_TWO_NAME))
        activity_header = [START_TIME, END_TIME, ACTIVITY]

        assert (df_one.columns == activity_header).any()
        assert (df_two.columns == activity_header).any()
        assert len(df_one) == 0
        assert len(df_two) == 0

    def test_devices_loaded(self):
        df_devices = self.data.df_devices
        device_header = [TIME, DEVICE, VAL]

        assert (df_devices.columns == device_header).all()
        assert len(df_devices) == 0

    def test_stats_activities(self):
        from pyadlml.dataset.stats.activities import activities_dist, activities_count, \
            activities_transitions, activity_duration
        df = self.data.df_activities_one
        lst = self.data.lst_activities

        act_count = activities_count(df, lst)
        assert len(act_count) == len(lst)

        act_trans = activities_transitions(df, lst)
        assert len(act_trans) == len(lst)
        assert act_trans.values.sum() == 0.0

        act_durs = activity_duration(df, lst)
        assert len(act_durs) == len(lst)
        assert act_durs['minutes'].values.sum() == 0.0

        act_dist = activities_dist(df, lst, n=100)
        assert len(act_dist.columns) == len(lst)
        assert len(act_dist) == 100


    def test_plot_activities(self):
        from pyadlml.dataset.plot.activities import boxplot, \
            transitions, total_duration, density_one_day, counts

        df = self.data.df_activities_one
        lst = self.data.lst_activities

        counts(df)
        counts(df, lst_acts=lst)

        boxplot(df)
        boxplot(df, lst_acts=lst)

        try:
            total_duration(df)
        except ValueError:
            pass
        total_duration(df, lst_acts=lst)

        transitions(df)
        transitions(df, lst_acts=lst)

        try:
            density_one_day(df)
        except ValueError:
            pass
        density_one_day(df, lst_acts=lst, n=10)

class TestPartialDataset(unittest.TestCase):
    def setUp(self):
        dataset_dir = str(script_directory) + '/datasets/partial_dataset'
        self.data = load_act_assist(dataset_dir, subjects=[SUBJECT_ADMIN_NAME])

    def test_activities_loaded(self):
        df = getattr(self.data, 'df_activities_{}'.format(SUBJECT_ADMIN_NAME))
        activity_header = [START_TIME, END_TIME, ACTIVITY]

        assert (df.columns == activity_header).any()
        assert len(df) == 3

    def test_devices_loaded(self):
        df_devices = self.data.df_devices
        device_header = [TIME, DEVICE, VAL]

        assert (df_devices.columns == device_header).all()
        assert len(df_devices) == 12

    def test_stats_activities(self):
        from pyadlml.dataset.stats.activities import activities_dist, activities_count, \
            activities_transitions, activity_duration
        df = self.data.df_activities_admin
        lst = self.data.lst_activities

        act_count = activities_count(df, lst)
        assert len(act_count) == len(lst)
        act_count = activities_count(df)
        assert len(act_count) == 2

        act_trans = activities_transitions(df, lst)
        assert len(act_trans) == len(lst)
        assert act_trans.values.sum() == len(df) - 1
        act_trans = activities_transitions(df)
        assert len(act_trans) == 2

        act_durs = activity_duration(df, lst)
        assert len(act_durs) == len(lst)
        act_durs = activity_duration(df)
        assert len(act_durs) == 2

        act_dist = activities_dist(df, lst, n=100)
        assert len(act_dist.columns) == len(lst)
        assert len(act_dist) == 100
        act_dist = activities_dist(df, n=100)
        assert len(act_dist.columns) == 2
        assert len(act_dist) == 100

    def test_stats_devices(self):
        from pyadlml.dataset.stats.devices import event_cross_correlogram, event_count, \
            events_one_day, inter_event_intervals, on_off_stats, state_cross_correlation

        df = self.data.df_devices
        lst = self.data.lst_devices
        recorded_devs = 5 # only those devices have an entry in devices.csv

        #tcorr = device_tcorr(df)
        tcorr = event_cross_correlogram(df, lst)
        assert tcorr.shape[0] == len(lst) and tcorr.shape[1] == len(lst)

        trigg_c = event_count(df)
        assert len(trigg_c) == recorded_devs
        trigg_c = event_count(df, lst)
        assert len(trigg_c) == len(lst)

        trigg_od = events_one_day(df)
        assert len(trigg_od.columns) == recorded_devs
        trigg_od = events_one_day(df, lst)
        assert len(trigg_od.columns) == len(lst)

        trigg_td = inter_event_intervals(df)
        assert len(trigg_td) == len(df) - 1

        onoff = on_off_stats(df)
        assert len(onoff) == recorded_devs
        onoff = on_off_stats(df, lst)
        assert len(onoff) == len(lst)

        dc = state_cross_correlation(df)
        assert len(dc) == recorded_devs
        dc = state_cross_correlation(df, lst)
        assert len(dc) == len(lst)


    def test_plot_activities(self):
        from pyadlml.dataset.plot.activities import activities_duration_dist, boxplot, \
            transitions, total_duration, density_one_day, counts

        df = self.data.df_activities_admin
        lst = self.data.lst_activities

        counts(df)
        counts(df, lst_acts=lst)

        boxplot(df)
        boxplot(df, lst_acts=lst)

        total_duration(df)
        total_duration(df, lst_acts=lst)

        transitions(df)
        transitions(df, lst_acts=lst)

        density_one_day(df)
        density_one_day(df, lst_acts=lst, n=10)

    def test_plot_devices(self):
        from pyadlml.dataset.plot.devices import event_count, event_density_one_day, states_cross_correlation,\
           state_fractions, state_boxplot, state_cross_correlation, inter_event_intervals

        df = self.data.df_devices
        lst = self.data.lst_devices

        event_count(df)
        event_count(df_devs=df, lst_devs=lst)

        event_density_one_day(df)
        event_density_one_day(df, lst_devs=lst)

        states_cross_correlation(df)
        states_cross_correlation(df, lst_devs=lst)

        inter_event_intervals(df)

        state_fractions(df)
        state_fractions(df, lst_devs=lst)

        state_boxplot(df)
        state_boxplot(df, lst_devs=lst)

        state_cross_correlation(df)
        state_cross_correlation(df, lst_devs=lst)


if __name__ == '__main__':
    unittest.main()


import pandas as pd
import numpy as np
from pyadlml.constants import START_TIME, END_TIME, ACTIVITY, TIME
from pyadlml.dataset._core.activities import add_other_activity, ActivityDict
from pyadlml.dataset.stats.util import df_density_binned
from datetime import timezone, datetime


def activity_order_by_duration(df_acts: pd.DataFrame) -> list:
    """ Computes a list of activities ordered by the cumulated time per activity

    Parameters
    ----------
    df_acts : pd.DataFrame
        An activity dataframe with columns 'start_time', 'end_time' and 'activity'.

    Returns
    -------
    list
        An ordered list where the first is the most prominent activity, followed by the second most prominent ...
    """
    if isinstance(df_acts, pd.DataFrame):
        dct_acts = ActivityDict.wrap(df_acts)
    else: 
        dct_acts = df_acts
    dfs = [activity_duration(df).set_index(ACTIVITY) for df in dct_acts.values()]

    df = pd.concat(dfs, axis=1).fillna(0)
    df['sum'] = df.sum(axis=1)
    df = df.sort_values(by='sum', ascending=False)
    return df.index.to_list()


def activity_order_by_count(df_acts: pd.DataFrame) -> list:
    """ Computes a list ordered by the most cumulated time per activity

    Parameters
    ----------
    df_acts : pd.DataFrame
        An activity dataframe with columns 'start_time', 'end_time' and 'activity'.

    Returns
    -------
    list
        An ordered list where the first is the most prominent activity, followed by the second most prominent ...
    """
    if isinstance(df_acts, pd.DataFrame):
        dct_acts = ActivityDict.wrap(df_acts)
    else: 
        dct_acts = df_acts
    dfs = [activities_count(df).set_index(ACTIVITY) for df in dct_acts.values()]

    df = pd.concat(dfs, axis=1).fillna(0)
    df['sum'] = df.sum(axis=1)
    df = df.sort_values(by='sum', ascending=False)
    return df.index.to_list()


def _get_freq_func(freq):
    """ returns the correct transform function of time differences into integers
        the column on which the function is applied has to be of type timedelta
    Parameters
    ----------


    Returns
    -------
    """
    assert freq in ['minutes', 'hours', 'seconds', 'm', 'h', 's']
    if freq == 'seconds' or freq == 's':
        return lambda x: x.total_seconds()
    elif freq == 'minutes' or freq == 'm':
        return lambda x: x.total_seconds()/60
    else:
        return lambda x: x.total_seconds()/3600

def coverage(df_activities, df_devices, datapoints=False):
    """ Computes the activity coverage for the devices.

    Parameters
    ----------
    df_activities : pd.DataFrame
        Todo
    datapoints: bool, default=False
        Determines whether the activity coverage is computed with respect
        to the datapoints covered or the total time.

    Returns
    -------
    cov : float
        The percentage covered by activities
    """
    if datapoints:
        from pyadlml.dataset._core.acts_and_devs import label_data
        df_acts = df_activities.copy()
        lbl_devs = label_data(df_devices, df_acts)
        nr_nans = lbl_devs[ACTIVITY].isna().sum()
        return 1 - nr_nans/len(lbl_devs)
    else:
        durations = activity_duration(df_activities, idle=True)
        total = durations['minutes'].sum()
        amount_idle = durations.loc[(durations[ACTIVITY] == 'idle'), 'minutes'].values[0]
        return 1 - amount_idle/total

def activities_duration_dist(df_acts, lst_acts=None, freq='minutes', idle=False):
    """ Computes the activity distribution.

    Parameters
    ----------
    df_acts : pd.DataFrame
        All recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    lst_acts : list, optional
        A list of activities that are included in the statistic. The list can be a
        subset of the recorded activities or contain activities that are not recorded.
    freq : str, optional
        Defaults to 1000.
        Can be one of 'minutes', 'seconds', 'hours', 'm', 'h', 's'.
    idle : bool, default=False
        Whether to include the idle activity. More in user guide

    Returns
    -------
    df : pd.DataFrame
        First column activity names second column total time in frequency.

    Example
    -------
    .. code python::
        TODO

    """
    df = df_acts.copy()
    if idle:
        df = add_other_activity(df)

    # integrate the time difference for activities
    diff = 'total_time'
    df[diff] = df[END_TIME] - df[START_TIME]
    df = df[[ACTIVITY, diff]]

    # if no data is logged for a certain activity append 0.0 duration value
    #if lst_acts is not None:
    #    for activity in set(lst_acts).difference(set(list(df[ACTIVITY]))):
    #        df = df.append(pd.DataFrame(
    #            data=[[activity, pd.Timedelta('0ns')]],
    #            columns=df.columns, index=[len(df)]))

    df[freq] = df[diff].apply(_get_freq_func(freq))
    return df.sort_values(by=ACTIVITY)

def activity_duration(df_acts, time_unit='minutes', normalize=False, idle=False):
    """ Compute how much time an inhabitant spent performing an activity

    Parameters
    ----------
    df_acts : pd.DataFrame
        All recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    time_unit : str of {'minutes', 'seconds', 'hours', 'm', 's', 'h'}
        The unit of time the durations are returned in.
    normalize : bool, default=False
        If set to *true* rather than returning the durations in a specified
        time unit, the fraction of the durations is returned.

    Returns
    -------
    df : pd.Dataframe

    Examples
    --------
    >>> from pyadlml.dataset import fetch_amsterdam
    >>> from pyadlml.stats import activity_duration
    >>> data = fetch_amsterdam()
    >>> activity_duration(data.df_activities, time_unit='m')
                activity       minutes
    0          get drink     16.700000
    1          go to bed  11070.166267
    2        leave house  22169.883333
    3  prepare Breakfast     63.500000
    4     prepare Dinner    338.899967
    5        take shower    209.566667
    6         use toilet    195.249567

    """
    df = activities_duration_dist(df_acts, freq=time_unit, idle=idle)
    if df.empty:
        raise ValueError("no activity was recorded")
    df = df.groupby(ACTIVITY).sum().reset_index()
    df.columns = [ACTIVITY, time_unit]

    # compute fractions of activities to each other
    if normalize:
        normalize = df[time_unit].sum()
        df[time_unit] = df[time_unit].apply(lambda x: x / normalize)

    return df.sort_values(by=ACTIVITY)

def activities_count(df_acts: pd.DataFrame) -> pd.DataFrame:
    """ Computes how many times a certain activity occurs within the dataset.

    Parameters
    ----------
    df_acts : pd.DataFrame
        All recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.

    Returns
    -------
    pd.Dataframe

    Examples
    --------
        >>> from pyadlml.dataset import fetch_amsterdam
        >>> from pyadlml.stats import activity_count
        >>> data = fetch_amsterdam()
        >>> activity_count(data.df_activities)
                    activity  occurrence
        0          get drink         19
        1          go to bed         47
        2        leave house         33
        3  prepare Breakfast         19
        4     prepare Dinner         12
        5        take shower         22
        6         use toilet        111
    """
    res_col_name = 'occurrence'
    df = df_acts.copy()

    # Count the occurences of the activity
    df = df.groupby(ACTIVITY).count().reset_index()
    df = df.drop(columns=[END_TIME])
    df.columns = [ACTIVITY, res_col_name]

    # correct for missing activities
    #if lst_acts is not None:
    #    diff = set(lst_acts).difference(set(list(df[ACTIVITY])))
    #    for activity in diff:
    #        df = df.append(pd.DataFrame(data=[[activity, 0.0]], columns=df.columns, index=[len(df) + 1]))

    return df.sort_values(by=ACTIVITY)


def activities_transitions(df_acts):
    """  Compute a transition matrix that displays how often one
    activity was followed by another.

    Parameters
    ----------
    df_acts : pd.DataFrame
        All recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.

    Returns
    -------
    df : pd.Dataframe

    Notes
    -----
    The transition frequency :math:`c_{ij}` describes how often
    an activity :math:`i` was followed by activity :math:`j`.

    Examples
    --------
        >>> from pyadlml.dataset import fetch_amsterdam
        >>> from pyadlml.stats import activity_transition
        >>> data = fetch_amsterdam()
        >>> activity_transition(data.df_activities)
        act_after          get drink  go to bed  ...  use toilet
        activity
        get drink                  3          0  ...          15
        go to bed                  0          0  ...          43
        leave house                3          1  ...          22
        prepare Breakfast          1          0  ...           8
        prepare Dinner             7          0  ...           4
        take shower                0          0  ...           1
        use toilet                 5         46  ...          18
    """

    df = df_acts.copy()
    df = df[[ACTIVITY]]
    df['act_after'] = df[ACTIVITY].shift(-1)
    df = pd.crosstab(df[ACTIVITY], df['act_after'])

    cols = list(df.columns)
    rows = list(df.index)
    for r in rows:
        if r not in cols:
            break

    # Catches the case when either an activity is only followed once and not preceeded by another
    # activity or an activity is only preceeded once and not succeded by another activity
    if len(df.columns) != len(df.index):
        row_diff = set(df.columns).difference(set(list(df.index)))
        col_diff = set(df.index).difference(set(list(df.columns)))
        for activity in col_diff:
            # when columns are missing
            df[activity] = 0
        for activity in row_diff:
            # when rows are missing
            df = df.append(pd.DataFrame(data=0.0, columns=df.columns, index=[activity]))

    # sort activities alphabetically
    df = df.sort_index(axis=0)
    df = df.sort_index(axis=1)

    return df


def activities_dist(df_acts, n=1000, dt=None, relative=False):
    """
    Approximate the activity densities for one day by
    using monte-carlo sampling from the activity intervals.

    Parameters
    ----------
    df_acts : pd.DataFrame
        All recorded activities from a dataset. Fore more information refer to the
        :ref:`user guide<activity_dataframe>`.
    n : int, optional
        The number of samples to draw from each activity. Defaults to 1000.
    dt : str of { 'xm', }, default=None
        if set

    Examples
    --------
    >>> from pyadlml.stats import activity_dist
    >>> transitions = activity_dist(data.df_activities, n=2000)
              prepare Dinner           get drink ...         leave house
    0    1990-01-01 18:12:39 1990-01-01 21:14:07 ... 1990-01-01 13:30:33
    1    1990-01-01 20:15:14 1990-01-01 20:23:31 ... 1990-01-01 12:03:13
    ..                      ...                 ...                 ...
    1999 1990-01-01 18:16:27 1990-01-01 08:49:38 ... 1990-01-01 16:18:25

    Returns
    -------
    df : pd.Dataframe
        Each row represents density point.
    """

    # todo
    #        principle: ancestral sampling
    #        select an interval at random
    #        sample uniform value of the interval
    #        sample from gaussian dist centered at interval instead
    #        of uniform dist
    #        solves the problem that beyond interval limits there can't be
    #        any sample
    # Todo include range for day or week

    df = df_acts.copy()
    activities = df_acts[ACTIVITY].unique()

    #res = pd.DataFrame(index=range(n), columns=activities, data=0)
    res = pd.DataFrame(index=range(len(activities)*n), columns=[TIME, ACTIVITY], dtype=object)

    # sample for each column separately
    for i, activity in enumerate(activities):
        series = _sample_ts(df.loc[df[ACTIVITY] == activity, [START_TIME, END_TIME]], n)
        res.loc[n*i:n*i+n-1, TIME] = series.values
        res.loc[n*i:n*i+n-1, ACTIVITY] = activity

    if relative and dt is None:
        res[TIME] = pd.to_datetime('1/1/2000 - ' + res[TIME])
        res[TIME] = res[TIME] - np.datetime64("2000-01-01 00:00:00")
        res[TIME] = res[TIME] / np.timedelta64(1, 's')

    if dt is not None:
        res[TIME] = pd.to_datetime('1/1/2000 - ' + res[TIME])
        res = df_density_binned(res, column_str=ACTIVITY, dt=dt)

    return res
    



def _sample_ts(df, n):
    """ Samples one column of activities. First chooses one activity at random and then samples a timepoint from
        within that activities interval

    Parameters
    ----------
    df: pd.DataFrame
            dataframe with two time intervals
            | start_time    | end_time  | .... |
            ------------------------------------
            | ts            | ts        | ...

    Returns
    -------
    pd.Series
        The sampled timestamps
    """
    assert START_TIME in df.columns and END_TIME in df.columns

    xs = np.empty(shape=(n,), dtype=object)
    for i in range(n):
        interval = df.iloc[np.random.randint(0, df.shape[0])]

        # convert to unix timestamp
        unx_min = interval.start_time.replace(tzinfo=timezone.utc).timestamp()
        unx_max = interval.end_time.replace(tzinfo=timezone.utc).timestamp()

        # sample uniform and convert back to timestamp
        try:
            unx_sample = np.random.randint(unx_min, unx_max)
        except ValueError:
            if unx_min == unx_max:
                unx_sample = unx_min
            else:
                raise ValueError('Sampling interval went wrong')
        xs[i] = datetime.utcfromtimestamp(unx_sample).strftime('%H:%M:%S')

    return pd.Series(xs)



from pyadlml.constants import ACTIVITY, BOOL, CAT, START_TIME, END_TIME, STRFTIME_DATE
import numpy as np
import plotly.graph_objects as go
import plotly.express as px
import dash
from dash.exceptions import PreventUpdate
import bisect
import pandas as pd


def remove_whitespace_around_fig(plot_func):
    def wrapper(*args, **kwargs):
        fig = plot_func(*args, **kwargs)
        fig.update_layout(margin=dict(l=0, r=0, b=0, t=30, pad=0))
        return fig
    return wrapper

def _style_colorbar(fig, label):
    fig.update_traces(colorbar=dict(
        title=dict(text=label, font=dict(size=10)),
        titleside='right',
        thickness=10,
    ))

def _dyn_y_label_size(plot_height, nr_devs):
    """Maps bar heights to visually pleasing size for different nr. of labels"""
    mp = np.array([[15, 9],     # val < 15  -> 9
                   [20, 11],
                   [30, 11],
                   [40, 11],
                   [80, 11],
                  [100, 11],
    ])
    return range2value(mp, nr_devs)


def _dyn_marker_size(plot_height, nr_devs):
    """Maps bar heights to visually pleasing size for different nr. of labels"""
    mp = np.array([[15, 3],  # val < 15 -> 3
                   [50, 5],
                  [100, 10]])
    return range2value(mp, nr_devs)

def range2value(mp, val):
    index = min(max(0, bisect.bisect_right(mp[:,0], val)-1), len(mp[:,0])-1)
    return mp[:,1][index]

import pandas as pd
import json



def plot_histogram(fig, row, col, bins, counts, dev1, dev2, cm):
    """
    Plot a histogram into a figure  
    """
    counts = np.concatenate([counts[:int(len(counts)/2)],
                        counts[int(len(counts)/2)+1:]]) 

    tmp = pd.interval_range(start=pd.Timedelta('0s'), end=pd.Timedelta(max_lag), freq=pd.Timedelta(binsize)).values
    len_cd = len(bins)
    cd = np.full(bins.shape, fill_value='', dtype=object)
    for i in range(0, len(cd)):
        if i < len(tmp)-1:
            cd[i] = '(-' + str(tmp[len(tmp)-i-1].right) + ', -' + str(tmp[len(tmp)-i-1].left) + ']'
        elif i == len(tmp)-1:
            cd[i] = '(-' + str(tmp[len(tmp)-i-1].right) + ', ' + str(tmp[len(tmp)-i-1].left) + ']'
        elif i == len(tmp):
            cd[i] = '(0 days 00:00:00, 0 days 00:00:00]'
        else:
            cd[i] = str(tmp[(i-1)%len(tmp)])
    half = int(np.ceil(len(bins)/2))
    before = np.full(half, fill_value='before', dtype=object)
    after = np.full(half, fill_value='after', dtype=object)
    cd1 = np.concatenate([before, after])[:-1]
    cd1[half-1] = 'at'
    cd = np.stack([cd, cd1]).T 


    fig.add_trace(
        go.Bar(
            x=bins,
            y=counts,
            opacity=1,
            customdata=cd, 
            hovertemplate='%{y} events of "' + dev1 + '" occur during %{customdata[0]}<br>%{customdata[1]}' + f' a type "{dev2}" event<extra></extra>',
            showlegend=False,
            marker=dict(line_width=0, color=cm[dev2])
        ), row=row, col=col
    )
    fig.update_xaxes(showticklabels=False, row=row, col=col)
    fig.update_yaxes(showticklabels=False, row=row, col=col)

    return fig





def format_device_labels(labels: list, dtypes: dict, order='alphabetical',
                         boolean_state=False, categorical_state=False, custom_rule=None):
    """ Sorts devices after a given rule. Format the devices labels and produces
    an ordering to sort values given the labels.

    Parameters
    ----------
    labels : nd array
        List of labels, where boolean devices have the format [dev_bool_x:on, dev_bool_x:off'] and
        categorical devices where format [dev_cat:subcat1, ..., dev_cat:sub_cat3]
    dtypes : dict
        The dtypes of the devices
    order : str of {'alphabetical', 'areas', 'custom'}, default='alphabetical'
        The criteria with which the devices are sorted.
    boolean_state : bool, default=False
        Indicates whether the boolean devices are split into

    custom_rule : func, default=None


    Returns
    -------
    lst : list
        Result list of correctly formatted device names
    new_order : list
        The indices that have to be reordered
    """
    assert order in ['alphabetical', 'areas', 'custom']
    DELIM = ':'
    ON = ':on'
    OFF = ':off'
    if isinstance(labels, list):
        labels = np.array(labels, dtype='object')

    def format_dev_and_state(word):
        return ''.join(' - ' if c == DELIM else c for c in word)

    def only_state(word):
        return word.split(DELIM)[1]

    def only_dev(word):
        return word.split(DELIM)[0]

    # presort devices
    devices = np.concatenate(list(dtypes.values()))
    if order == 'alphabetical' and custom_rule is None:
        devices = np.sort(devices)
    elif order == 'areas' and custom_rule is None:
        raise NotImplementedError
    elif order == 'custom' and custom_rule is not None:
        devices = custom_rule(devices)

    # rename devices
    new_labels = np.zeros((len(labels)), dtype=object)
    new_order = np.zeros((len(labels)), dtype=np.int32)

    dev_idx = 0
    i = 0
    while i < len(new_labels):
        dev = devices[dev_idx]
        dev_idx += 1
        if boolean_state and dev in dtypes[BOOL]:
            idx_on = np.where(labels == (dev + ON))[0][0]
            idx_off = np.where(labels == (dev + OFF))[0][0]

            new_labels[i] = format_dev_and_state(labels[idx_off])
            new_labels[i+1] = only_state(labels[idx_on])
            new_order[i] = idx_off
            new_order[i+1] = idx_on
            i += 2
        elif categorical_state and dev in dtypes[CAT]:
            mask = [lbl.split(DELIM)[0] == dev for lbl in labels]
            idxs = np.where(mask)[0]
            cats = labels[mask]
            cats_new_order = np.argsort(cats)
            for j in range(len(cats)):
                if j == 0:
                    new_labels[i] = format_dev_and_state(cats[cats_new_order[j]])
                else:
                    new_labels[i] = "- " + only_state(cats[cats_new_order[j]])
                new_order[i] = idxs[cats_new_order[j]]
                i += 1
        else:
            idx = np.where(labels == dev)[0]
            new_labels[i] = dev
            new_order[i] = idx
            i += 1

    return new_labels, new_order



def plot_activities_into(fig, df_acts: pd.DataFrame, y_label: str, cat_col_map, 
                         row=1, col=1, act_order=None
    ):


    fig.update_xaxes(type="date")

    # RLE
    df = df_acts.copy()

    df = df.sort_values(by=START_TIME).reset_index(drop=True)
    df['y_label'] = y_label
    df = df[['y_label', START_TIME, END_TIME, ACTIVITY]]

    df['dur'] = (df[END_TIME] - df[START_TIME])
    df['lengths' ] = df['dur'].astype("timedelta64[ms]")
    df['dur'] = df['dur'].astype(str)


    if act_order is None:
        act_order = df[ACTIVITY].uniques()

    trace_lst = []
    set_legendgroup_title = 'Activities'
    for act_name in act_order:
        cat_col_map.update(act_name, fig)
        show_legend = act_name not in legend_current_items(fig)
        act_mask = (df[ACTIVITY] == act_name)
        hover_template = '<b>' + act_name + '</b><br>'\
                        + 'Start_time: %{base|' + STRFTIME_DATE + '}<br>' \
                        + 'End_time: %{x| ' + STRFTIME_DATE + '}<br>' \
                        + 'Dur: %{customdata}<extra></extra>'

        trace = go.Bar(name=act_name,
                        meta=y_label,
                        base=df.loc[act_mask, 'start_time'],
                        x=df.loc[act_mask, 'lengths'],
                        y=df.loc[act_mask, 'y_label'],
                        legendgrouptitle_text=set_legendgroup_title,
                        marker_color=cat_col_map[act_name],
                        legendgroup=act_name,
                        customdata=df.loc[act_mask, 'dur'],
                        orientation='h',
                        width=0.9,
                        textposition='auto',
                        alignmentgroup=True,
                        offsetgroup=act_name,
                        showlegend=show_legend,
                        hovertemplate=hover_template,
        )
        trace_lst.append(trace)
        set_legendgroup_title = None

    [fig.add_trace(trace, row=row, col=col) for trace in trace_lst]

    # Add additional 
    acts_not_in_legend = set(act_order) - set(legend_current_items(fig))
    for act_name in acts_not_in_legend:
        # TODO check if it works
        raise NotImplementedError
        cat_col_map.update(act_name, fig)
        fig.add_trace(go.Bar(
            name=act_name, 
            visible='legendonly',
            showlegend=True
        ))

    # Add traces, layout and frames to figure
    fig.update_layout({'barmode': 'overlay',
                    'legend': {'tracegroupgap': 0}
                    })

    return fig



def legend_current_items(fig):
    legend_current_items = [] 
    for trace in fig.data:
        if isinstance(trace, go.Bar) \
          and trace.legendgroup is not None:
            legend_current_items.append(trace.legendgroup)
    return legend_current_items


class CatColMap():
    COL_ON = 'teal'
    COL_OFF = 'lightgray'

    cat_idx =0

    cat_col_map = {
        0:COL_OFF, 1:COL_ON,
        'off':COL_OFF, 'on':COL_ON,
        False:COL_OFF, True:COL_ON,
    }
    def __init__(self, theme='pastel'):
        if theme == 'pastel':
            self.cat_colors = px.colors.qualitative.Pastel \
                    + px.colors.qualitative.Pastel1 \
                    + px.colors.qualitative.Pastel2
        elif theme == 'set':
            self.cat_colors = px.colors.qualitative.Set1 \
                    + px.colors.qualitative.Set2 \
                    + px.colors.qualitative.Set3
        else:
            self.cat_colors = px.colors.qualitative.T10

    def update(self, cat, fig=None):
        if fig is not None:
            if cat not in legend_current_items(fig):
                if cat not in self.cat_col_map.keys():
                    self.cat_col_map[cat] = self.cat_colors[self.cat_idx]
                    self.cat_idx +=1
                return True
            return False
        else:
            if cat not in self.cat_col_map.keys():
                self.cat_col_map[cat] = self.cat_colors[self.cat_idx]
                self.cat_idx +=1
            
    def __getitem__(self, sub):
        return self.cat_col_map[sub]

    def __setitem__(self, sub, item):
        self.cat_col_map[sub] = item

def dash_get_trigger_value(ctx=None):
    try:
        ctx = dash.callback_context if ctx is None else ctx
        return ctx.triggered[0]['value']
    except IndexError:
        raise PreventUpdate

def dash_get_trigger_element(ctx=None):
    try:
        ctx = dash.callback_context if ctx is None else ctx
        return ctx.triggered[0]['prop_id'].split('.')[0]
    except IndexError:
        raise PreventUpdate


deserialize_range = lambda x: [pd.Timestamp(ts) for ts in json.loads(x)] 
serialize_range = lambda x: json.dumps([ts.isoformat() for ts in x])
range_from_fig = lambda f: [pd.Timestamp(ts) for ts in f['layout']['xaxis']['range']]




import functools
import os
from collections import OrderedDict
from pathlib import Path
import inspect

import joblib
import hashlib
import zipfile
import shutil

import pandas as pd

from pyadlml.constants import ACTIVITY, DEVICE
from pyadlml.dataset._core.activities import correct_activities
from pyadlml.dataset._core.devices import correct_devices, CORRECTION_TS, CORRECTION_ONOFF_INCONS
from pyadlml.dataset.util import extract_kwargs


# TODO flag for deletion
#def correct_acts(func):
#    @functools.wraps(func)
#    def wrapper_correction(*args, **kwargs):
#        data = func(*args, **kwargs)
#        retain_corrections = kwargs['retain_corrections'] if 'retain_corrections' in kwargs else False
#
#        df_acts = data.get_activity_dict()
#        for key, df_act in df_acts.items():
#            df_act, correction_act = correct_activities(df_act, retain_corrections=retain_corrections)
#            data.set_activity_df(df_act, key)
#            if retain_corrections:
#                data.set_activity_correction(correction_act, key)
#        return data
#    return wrapper_correction
#
#def correct_devs(func):
#    @functools.wraps(func)
#    def wrapper_correction(*args, **kwargs):
#        data = func(*args, **kwargs)
#        retain_corrections = kwargs['retain_corrections'] if 'retain_corrections' in kwargs else False
#
#        df_dev, correction_dev_dict = correct_devices(data.df_devices, retain_correction=retain_corrections)
#        data.set_device_df(df_dev)
#
#        if retain_corrections:
#            data.set_device_correction(correction_dev_dict[CORRECTION_TS],
#                                       correction_dev_dict[CORRECTION_ONOFF_INCONS])
#
#        return data
#    return wrapper_correction
#
#
#
#def correct_acts_and_devs(func):
#    """ Wraps a function that returns a data object with activity, device dataframes
#        and applies corrections on to device and activity dataframes.
#
#    """
#    @functools.wraps(func)
#    def wrapper_correction(*args, **kwargs):
#        data = func(*args, **kwargs)
#        df_dev = data.df_devices
#        df_acts = data.get_activity_dict()
#
#        try:
#            retain_corrections = kwargs['retain_corrections']
#        except:
#            retain_corrections = False
#
#        for key, df_act in df_acts.items():
#            df_act, correction_act = correct_activities(df_act, retain_corrections=retain_corrections)
#            data.set_activity_df(df_act, key)
#            if retain_corrections:
#                data.set_activity_correction(correction_act, key)
#        df_dev, correction_dev_dict = correct_devices(df_dev, retain_correction=retain_corrections)
#        data.set_device_df(df_dev)
#
#        if retain_corrections:
#            data.set_device_correction(correction_dev_dict[CORRECTION_TS],
#                                       correction_dev_dict[CORRECTION_ONOFF_INCONS])
#
#        return data
#    return wrapper_correction





#def fetcher(ds_name, func_download, file_name, url, fn_cleaned=None, url_cleaned=None):
#    """ The fetcher manages downloading and caching of datasets and their cleaned verions.
#
#    The folder structure for a dataset is
#    datahome
#    |- dataset_name
#        |- original
#            |- data files
#            |- ...
#        cached.joblib
#        cleaned.joblib
#
#    Parameters
#    ----------
#    ds_name : str
#        The datasetname. Is also used for the folder
#    func_download : function
#        The function that downloads the dataset
#    file_name : name
#        Arguments that are used are passed to the downloader
#    """
#    FN_CLEANED = 'cleaned.joblib'
#    FN_CACHED = 'cached.joblib'
#    def decorator_fetcher(func):
#        @extract_kwargs
#        @functools.wraps(func)
#        def wrapper_fetcher(*args, **kwargs):
#
#            # Resolve folders
#            data_home = get_data_home()
#            _ensure_dh_folder_exists()
#            dataset_folder = Path(data_home).joinpath(ds_name)
#            original_folder = dataset_folder.joinpath('original')
#
#
#            # TODO refactor, critical get dataset selecting argument to set prefix for cache and cleaned version
#            sel = None
#            for a in kwargs.keys():
#                if a not in ['cache', 'keep_original', 'retain_corrections',
#                            'folder_path', 'load_cleaned']:
#                    sel = a
#            if sel is not None:
#                cached_name = FN_CACHED.split('.')[0] + '_' + kwargs[sel] + '.' + FN_CACHED.split('.')[1]
#                cleaned_name = FN_CLEANED.split('.')[0] + '_' + kwargs[sel] + '.' + FN_CLEANED.split('.')[1]
#                fp_cached_dataset = dataset_folder.joinpath(cached_name)
#                fp_cleaned_dataset = dataset_folder.joinpath(cleaned_name)
#            else:
#                fp_cached_dataset = dataset_folder.joinpath(FN_CACHED)
#                fp_cleaned_dataset = dataset_folder.joinpath(FN_CLEANED)
#
#
#            try:
#                cache = kwargs['cache']
#                keep_original = kwargs['keep_original']
#                folder_path = kwargs['folder_path']
#            except KeyError:
#                raise KeyError('The fetch_dataset signature is not set correctly.')
#            load_cleaned = kwargs.pop('load_cleaned', False)
#
#            #assert cache and folder_path is None, "Both parameters can not be set at the same time. Please specify only one."
#
#            # Only download data if it was not already fetched
#            # and no cached version exists that should be loaded
#            # and the load_cleaned flag is not set
#
#            if not (original_folder.exists() or (fp_cached_dataset.is_file() and cache)\
#                    or folder_path is not None) and not load_cleaned:
#                original_folder.mkdir(parents=True, exist_ok=True)
#                func_download(original_folder, file_name, url)
#            elif load_cleaned and not fp_cleaned_dataset.exists():
#                download_from_mega(dataset_folder, fn_cleaned, url_cleaned, unzip=False)
#                Path(dataset_folder).joinpath(fn_cleaned).rename(FN_CLEANED)
#
#            # Load from cached version if available and cache flag is set
#            # or load from cleaned version
#            # otherwise load from the function
#            if fp_cached_dataset.is_file() and cache and not load_cleaned:
#                data = joblib.load(fp_cached_dataset)
#            elif fp_cleaned_dataset.is_file() and load_cleaned and not cache:
#                data = joblib.load(fp_cleaned_dataset)
#            else:
#                kwargs['folder_path'] = str(original_folder)
#                data = func(*args, **kwargs)
#
#            # Save dataset if downloaded and not already cached
#            if cache and not fp_cached_dataset.is_file() and not load_cleaned:
#                joblib.dump(data, fp_cached_dataset)
#
#            # Clean up data
#            # if not cache and fp_cached_dataset.is_file():
#            # if not keep_original and original_folder.exists():
#            #    _delete_data(original_folder)
#
#            return data
#        return wrapper_fetcher
#    return decorator_fetcher
#
#
#def fetch_handler(keep_original, cache, dataset_name,
#        mega_filename, mega_url,
#        load_func, data_postfix=''):
#    """ handles the downloading, loading and caching of a dataset
#    Parameters
#    ----------
#    Returns
#    -------
#    data : object
#
#    """
#    data_home = get_data_home()
#    _ensure_dh_folder_exists()
#    data_home_dataset = ''.join([data_home, '/', dataset_name])
#    cache_data_folder = _data_2_folder_name(data_home, dataset_name)
#
#    # download data    
#    if not os.path.isdir(data_home_dataset):
#        # download file from mega # TODO make official way available
#        download_from_mega(get_data_home(), mega_filename, mega_url)
#
#    # load data
#    if data_postfix != '':
#        data_name = cache_data_folder + '/' \
#            + DATA_DUMP_NAME[:-7] + '_' + data_postfix + '.joblib'
#    else:
#        data_name = cache_data_folder + '/' + DATA_DUMP_NAME
#
#    if Path(data_name).is_file(): 
#        data = joblib.load(data_name) 
#    else:
#        data = load_func(data_home_dataset + '/')
#        if cache:
#            _create_folder(cache_data_folder)
#            joblib.dump(data, data_name)
#            Path(os.path.join(cache_data_folder,dataset_name)).touch()
#
#
#    # clean up data
#    # TODO note that the folder is deleted. For two different subjects
#    # caching one and deleting another leads to deletion of the first 
#    if not cache and os.path.exists(cache_data_folder):
#        _delete_data(cache_data_folder)
#    if not keep_original and os.path.exists(data_home_dataset):
#        _delete_data(data_home_dataset)
#
#    return data


